from typing import List


class Renderer:
    async def render_file(self, file_path: str) -> None:
        raise NotImplementedError

    async def render_tree(self, tree_path: str) -> None:
        raise NotImplementedError


class SequentialRenderer(Renderer):
    def __init__(self, renderers: List[Renderer]):
        self._renderers = renderers

    async def render_file(self, file_path: str) -> None:
        for renderer in self._renderers:
            await renderer.render_file(file_path)

    async def render_tree(self, tree_path: str) -> None:
        for renderer in self._renderers:
            await renderer.render_tree(tree_path)



import unittest
from json import dump
from os import path, makedirs
from tempfile import TemporaryDirectory
from typing import Callable, Dict
from unittest.mock import patch

import click
from click.testing import CliRunner

import betty
from betty import os
from betty.error import UserFacingError
from betty.plugin import Plugin
from betty.serve import Server
from betty.tests import patch_cache, TestCase

try:
    from unittest.mock import AsyncMock
except ImportError:
    from mock.mock import AsyncMock

from betty.cli import main, CommandProvider, global_command, catch_exceptions
from betty.site import Site


class TestCommandError(BaseException):
    pass


class TestPlugin(Plugin, CommandProvider):
    @property
    def commands(self) -> Dict[str, Callable]:
        return {
            'test': self._test_command,
        }

    @click.command(name='test')
    @global_command
    async def _test_command(self):
        raise TestCommandError


@patch('sys.stderr')
@patch('sys.stdout')
class MainTest(TestCase):
    def test_without_arguments(self, _, __):
        runner = CliRunner()
        result = runner.invoke(main, catch_exceptions=False)
        self.assertEqual(0, result.exit_code)

    def test_help_without_configuration(self, _, __):
        runner = CliRunner()
        result = runner.invoke(main, ('--help',), catch_exceptions=False)
        self.assertEqual(0, result.exit_code)

    def test_configuration_without_help(self, _, __):
        with TemporaryDirectory() as working_directory_path:
            configuration_file_path = path.join(working_directory_path, 'betty.json')
            with TemporaryDirectory() as output_directory_path:
                url = 'https://example.com'
                config_dict = {
                    'output': output_directory_path,
                    'base_url': url,
                }
                with open(configuration_file_path, 'w') as f:
                    dump(config_dict, f)

                runner = CliRunner()
                result = runner.invoke(main, ('-c', configuration_file_path), catch_exceptions=False)
                self.assertEqual(2, result.exit_code)

    def test_help_with_configuration(self, _, __):
        with TemporaryDirectory() as working_directory_path:
            configuration_file_path = path.join(working_directory_path, 'betty.json')
            with TemporaryDirectory() as output_directory_path:
                url = 'https://example.com'
                config_dict = {
                    'output': output_directory_path,
                    'base_url': url,
                    'plugins': {
                        TestPlugin.name(): None,
                    },
                }
                with open(configuration_file_path, 'w') as f:
                    dump(config_dict, f)

                runner = CliRunner()
                result = runner.invoke(main, ('-c', configuration_file_path, '--help',), catch_exceptions=False)
                self.assertEqual(0, result.exit_code)

    def test_help_with_invalid_configuration_file_path(self, _, __):
        with TemporaryDirectory() as working_directory_path:
            configuration_file_path = path.join(working_directory_path, 'non-existent-betty.json')

            runner = CliRunner()
            result = runner.invoke(main, ('-c', configuration_file_path, '--help',), catch_exceptions=False)
            self.assertEqual(1, result.exit_code)

    def test_help_with_invalid_configuration(self, _, __):
        with TemporaryDirectory() as working_directory_path:
            configuration_file_path = path.join(working_directory_path, 'betty.json')
            config_dict = {}
            with open(configuration_file_path, 'w') as f:
                dump(config_dict, f)

            runner = CliRunner()
            result = runner.invoke(main, ('-c', configuration_file_path, '--help',), catch_exceptions=False)
            self.assertEqual(1, result.exit_code)

    def test_with_discovered_configuration(self, _, __):
        with TemporaryDirectory() as betty_site_path:
            with TemporaryDirectory() as output_directory_path:
                with open(path.join(betty_site_path, 'betty.json'), 'w') as config_file:
                    url = 'https://example.com'
                    config_dict = {
                        'output': output_directory_path,
                        'base_url': url,
                        'plugins': {
                            TestPlugin.name(): None,
                        },
                    }
                    dump(config_dict, config_file)
                with os.ChDir(betty_site_path):
                    runner = CliRunner()
                    result = runner.invoke(main, ('test',), catch_exceptions=False)
                    self.assertEqual(1, result.exit_code)


class CatchExceptionsTest(unittest.TestCase):
    def test_logging_user_facing_error(self) -> None:
        error_message = 'Something went wrong!'
        with self.assertLogs() as watcher:
            with self.assertRaises(SystemExit):
                with catch_exceptions():
                    raise UserFacingError(error_message)
            self.assertEquals('ERROR:root:%s' % error_message, watcher.output[0])

    def test_logging_uncaught_exception(self) -> None:
        error_message = 'Something went wrong!'
        with self.assertLogs() as watcher:
            with self.assertRaises(SystemExit):
                with catch_exceptions():
                    raise Exception(error_message)
            self.assertTrue(watcher.output[0].startswith('ERROR:root:%s' % error_message))
            self.assertIn('Traceback', watcher.output[0])


class ClearCachesTest(TestCase):
    @patch_cache
    def test(self):
        cached_file_path = path.join(betty._CACHE_DIRECTORY_PATH, 'KeepMeAroundPlease')
        open(cached_file_path, 'w').close()
        runner = CliRunner()
        result = runner.invoke(main, ('clear-caches',), catch_exceptions=False)
        self.assertEqual(0, result.exit_code)
        with self.assertRaises(FileNotFoundError):
            open(cached_file_path)


class GenerateTest(TestCase):
    @patch('betty.generate.generate', new_callable=AsyncMock)
    @patch('betty.parse.parse', new_callable=AsyncMock)
    def test(self, m_parse, m_generate):
        with TemporaryDirectory() as working_directory_path:
            configuration_file_path = path.join(working_directory_path, 'betty.json')
            with TemporaryDirectory() as output_directory_path:
                url = 'https://example.com'
                config_dict = {
                    'output': output_directory_path,
                    'base_url': url,
                }
                with open(configuration_file_path, 'w') as f:
                    dump(config_dict, f)

                runner = CliRunner()
                result = runner.invoke(main, ('-c', configuration_file_path, 'generate',), catch_exceptions=False)
                self.assertEqual(0, result.exit_code)

                m_parse.assert_called_once()
                parse_args, parse_kwargs = m_parse.await_args
                self.assertEquals(1, len(parse_args))
                self.assertIsInstance(parse_args[0], Site)
                self.assertEquals({}, parse_kwargs)

                m_generate.assert_called_once()
                render_args, render_kwargs = m_generate.call_args
                self.assertEquals(1, len(render_args))
                self.assertIsInstance(render_args[0], Site)
                self.assertEquals({}, render_kwargs)


class _KeyboardInterruptedServer(Server):
    def __init__(self, *args, **kwargs):
        pass

    async def start(self) -> None:
        raise KeyboardInterrupt


class ServeTest(TestCase):
    @patch('betty.serve.SiteServer', new_callable=lambda: _KeyboardInterruptedServer)
    def test(self, m_server):
        with TemporaryDirectory() as working_directory_path:
            configuration_file_path = path.join(working_directory_path, 'betty.json')
            with TemporaryDirectory() as output_directory_path:
                www_directory_path = path.join(output_directory_path, 'www')
                makedirs(www_directory_path)
                url = 'https://example.com'
                config_dict = {
                    'output': output_directory_path,
                    'base_url': url,
                }
                with open(configuration_file_path, 'w') as f:
                    dump(config_dict, f)

                runner = CliRunner()
                result = runner.invoke(main, ('-c', configuration_file_path, 'serve',), catch_exceptions=False)
                self.assertEqual(0, result.exit_code)

    def test_without_www_directory_should_error(self):
        with TemporaryDirectory() as working_directory_path:
            configuration_file_path = path.join(working_directory_path, 'betty.json')
            with TemporaryDirectory() as output_directory_path:
                url = 'https://example.com'
                config_dict = {
                    'output': output_directory_path,
                    'base_url': url,
                }
                with open(configuration_file_path, 'w') as f:
                    dump(config_dict, f)

                runner = CliRunner()
                result = runner.invoke(main, ('-c', configuration_file_path, 'serve',), catch_exceptions=False)
                self.assertEqual(1, result.exit_code)



import contextlib
import logging
import threading
import webbrowser
from http.server import SimpleHTTPRequestHandler, HTTPServer
from io import StringIO
from typing import Iterable

from betty.error import UserFacingError
from betty.os import ChDir
from betty.site import Site

DEFAULT_PORT = 8000


class ServerNotStartedError(RuntimeError):
    pass


class OsError(UserFacingError, OSError):
    pass


class Server:
    async def start(self) -> None:
        """
        Starts the server.
        :return: The public URL.
        """
        pass

    async def stop(self) -> None:
        """
        Stops the server.
        """
        pass

    @property
    def public_url(self) -> str:
        raise NotImplementedError

    async def __aenter__(self) -> 'Server':
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()


class ServerProvider:
    @property
    def servers(self) -> Iterable[Server]:
        raise NotImplementedError


class SiteServer(Server):
    def __init__(self, site: Site):
        self._site = site
        self._server = None

    def _get_server(self) -> Server:
        servers = (server for plugin in self._site.plugins.values() if isinstance(plugin, ServerProvider) for server in plugin.servers)
        with contextlib.suppress(StopIteration):
            return next(servers)
        return BuiltinServer(self._site.configuration.www_directory_path)

    async def start(self) -> None:
        self._server = self._get_server()
        await self._server.start()
        logging.getLogger().info('Serving your site at %s...' % self.public_url)
        webbrowser.open_new_tab(self.public_url)

    @property
    def public_url(self) -> str:
        return self._server.public_url

    async def stop(self) -> None:
        await self._server.stop()


class BuiltinServer(Server):
    def __init__(self, www_directory_path: str):
        self._www_directory_path = www_directory_path
        self._http_server = None
        self._cwd = None
        self._port = None

    async def start(self) -> None:
        logging.getLogger().info('Starting Python\'s built-in web server...')
        for self._port in range(DEFAULT_PORT, 65535):
            with contextlib.suppress(OSError):
                self._http_server = HTTPServer(('', self._port), SimpleHTTPRequestHandler)
                break
        if self._http_server is None:
            raise OsError('Cannot find an available port to bind the web server to.')
        self._cwd = ChDir(self._www_directory_path).change()
        threading.Thread(target=self._serve).start()

    @property
    def public_url(self) -> str:
        if self._port is not None:
            return 'http://localhost:%d' % self._port
        raise ServerNotStartedError('Cannot determine the public URL if the server has not started yet.')

    def _serve(self):
        with contextlib.redirect_stderr(StringIO()):
            self._http_server.serve_forever()

    async def stop(self) -> None:
        self._http_server.shutdown()
        self._cwd.revert()



from typing import Hashable


class LockError:
    pass


class AcquiredError(LockError, RuntimeError):
    def __init__(self, resource: Hashable):
        super().__init__(f'Cannot acquire a lock for "{resource}", because it is locked already.')


class Locks:
    def __init__(self):
        self._locks = set()

    def acquire(self, resource: Hashable):
        if resource in self._locks:
            raise AcquiredError(resource)
        self._locks.add(resource)

    def release(self, resource: Hashable):
        self._locks.discard(resource)



import gettext
from os import makedirs, path
from tempfile import TemporaryDirectory
from typing import List, Optional

from parameterized import parameterized

from betty.locale import Localized, negotiate_localizeds, Date, format_datey, DateRange, Translations, negotiate_locale, \
    Datey, open_translations
from betty.tests import TestCase


class DateTest(TestCase):
    def test_year(self):
        year = 1970
        sut = Date(year=year)
        self.assertEquals(year, sut.year)

    def test_month(self):
        month = 1
        sut = Date(month=month)
        self.assertEquals(month, sut.month)

    def test_day(self):
        day = 1
        sut = Date(day=day)
        self.assertEquals(day, sut.day)

    def test_fuzzy(self):
        fuzzy = True
        sut = Date()
        sut.fuzzy = fuzzy
        self.assertEquals(fuzzy, sut.fuzzy)

    @parameterized.expand([
        (True, 1970, 1, 1),
        (False, None, 1, 1),
        (True, 1970, None, 1),
        (True, 1970, 1, None),
        (False, None, None, 1),
        (True, 1970, None, None),
        (False, None, None, None),
    ])
    def test_comparable(self, expected, year, month, day):
        sut = Date(year, month, day)
        self.assertEquals(expected, sut.comparable)

    @parameterized.expand([
        (True, 1970, 1, 1),
        (False, None, 1, 1),
        (False, 1970, None, 1),
        (False, 1970, 1, None),
        (False, None, None, 1),
        (False, 1970, None, None),
        (False, None, None, None),
    ])
    def test_complete(self, expected, year, month, day):
        sut = Date(year, month, day)
        self.assertEquals(expected, sut.complete)

    def test_to_range_when_incomparable_should_raise(self):
        with self.assertRaises(ValueError):
            Date(None, 1, 1).to_range()

    @parameterized.expand([
        (1970, 1, 1),
        (None, None, None),
    ])
    def test_parts(self, year, month, day):
        self.assertEquals((year, month, day), Date(year, month, day).parts)

    @parameterized.expand([
        (False, Date(1970, 2, 1)),
        (True, Date(1970, 2, 2)),
        (False, Date(1970, 2, 3)),
        (False, DateRange()),
    ])
    def test_in(self, expected, other):
        self.assertEquals(expected, other in Date(1970, 2, 2))

    @parameterized.expand([
        (False, Date(1970, 2, 1)),
        (False, Date(1970, 2, 2)),
        (True, Date(1970, 2, 3)),
        (False, Date(1970)),
        (False, Date(1970, 2)),
        (True, Date(1971)),
        (True, Date(1970, 3)),
    ])
    def test_lt(self, expected, other):
        self.assertEquals(expected, Date(1970, 2, 2) < other)

    @parameterized.expand([
        (True, Date(1970, 1, 1)),
        (False, Date(1970, 1, None)),
        (False, Date(1970, None, 1)),
        (False, Date(None, 1, 1)),
        (False, Date(1970, None, None)),
        (False, Date(None, 1, None)),
        (False, Date(None, None, 1)),
        (False, None),
    ])
    def test_eq(self, expected, other):
        self.assertEquals(expected, Date(1970, 1, 1) == other)
        self.assertEquals(expected, other == Date(1970, 1, 1))

    @parameterized.expand([
        (True, Date(1970, 2, 1)),
        (False, Date(1970, 2, 2)),
        (False, Date(1970, 2, 3)),
    ])
    def test_gt(self, expected, other):
        self.assertEquals(expected, Date(1970, 2, 2) > other)


class DateRangeTest(TestCase):
    _TEST_IN_PARAMETERS = [
        (False, Date(1970, 2, 2), DateRange()),
        (False, Date(1970, 2), DateRange()),
        (False, Date(1970), DateRange()),
        (False, Date(1970, 2, 1), DateRange(Date(1970, 2, 2))),
        (True, Date(1970, 2, 2), DateRange(Date(1970, 2, 2))),
        (True, Date(1970, 2, 3), DateRange(Date(1970, 2, 2))),
        (True, Date(1970, 2, 1), DateRange(None, Date(1970, 2, 2))),
        (True, Date(1970, 2, 2), DateRange(None, Date(1970, 2, 2))),
        (False, Date(1970, 2, 3), DateRange(None, Date(1970, 2, 2))),
        (False, Date(1969, 2, 1), DateRange(Date(1969, 2, 2), Date(1970, 2, 2))),
        (True, Date(1970, 2, 1), DateRange(Date(1969, 2, 2), Date(1970, 2, 2))),
        (False, Date(1971, 2, 1), DateRange(Date(1969, 2, 2), Date(1970, 2, 2))),
        (True, DateRange(Date(1970, 2, 1)), DateRange(Date(1970, 2, 2))),
        (True, DateRange(Date(1970, 2, 2)), DateRange(Date(1970, 2, 2))),
        (True, DateRange(Date(1970, 2, 3)), DateRange(Date(1970, 2, 2))),
        (False, DateRange(None, Date(1970, 2, 1)), DateRange(Date(1970, 2, 2))),
        (True, DateRange(None, Date(1970, 2, 2)), DateRange(Date(1970, 2, 2))),
        (True, DateRange(None, Date(1970, 2, 3)), DateRange(Date(1970, 2, 2))),
        (True, DateRange(Date(1970, 2, 1)), DateRange(None, Date(1970, 2, 2))),
        (True, DateRange(Date(1970, 2, 2)), DateRange(None, Date(1970, 2, 2))),
        (False, DateRange(Date(1970, 2, 3)), DateRange(None, Date(1970, 2, 2))),
        (True, DateRange(None, Date(1970, 2, 1)), DateRange(None, Date(1970, 2, 2))),
        (True, DateRange(None, Date(1970, 2, 2)), DateRange(None, Date(1970, 2, 2))),
        (True, DateRange(None, Date(1970, 2, 3)), DateRange(None, Date(1970, 2, 2))),
        (True, DateRange(Date(1969, 2, 1)), DateRange(Date(1969, 2, 2), Date(1970, 2, 2))),
        (True, DateRange(Date(1970, 2, 1)), DateRange(Date(1969, 2, 2), Date(1970, 2, 2))),
        (False, DateRange(Date(1971, 2, 1)), DateRange(Date(1969, 2, 2), Date(1970, 2, 2))),
        (False, DateRange(None, Date(1969, 2, 1)), DateRange(Date(1969, 2, 2), Date(1970, 2, 2))),
        (True, DateRange(None, Date(1970, 2, 1)), DateRange(Date(1969, 2, 2), Date(1970, 2, 2))),
        (True, DateRange(None, Date(1971, 2, 1)), DateRange(Date(1969, 2, 2), Date(1970, 2, 2))),
        (False, DateRange(Date(1969, 2, 2), Date(1970, 2, 2)), DateRange(Date(1971, 2, 2), Date(1972, 2, 2))),
        (True, DateRange(Date(1969, 2, 2), Date(1971, 2, 2)), DateRange(Date(1970, 2, 2), Date(1972, 2, 2))),
        (True, DateRange(Date(1970, 2, 2), Date(1971, 2, 2)), DateRange(Date(1969, 2, 2), Date(1972, 2, 2))),
    ]

    # Mirror the arguments because we want the containment check to work in either direction.
    @parameterized.expand(_TEST_IN_PARAMETERS + list(map(lambda x: (x[0], x[2], x[1]), _TEST_IN_PARAMETERS)))
    def test_in(self, expected: bool, other: Datey, sut: DateRange):
        self.assertEquals(expected, other in sut)

    @parameterized.expand([
        (False, Date(1970, 2, 1)),
        (False, Date(1970, 2, 2)),
        (True, Date(1970, 2, 3)),
        (False, DateRange(Date(1970, 2, 1))),
        (False, DateRange(Date(1970, 2, 2))),
        (True, DateRange(Date(1970, 2, 3))),
        (False, DateRange(None, Date(1970, 2, 1))),
        (False, DateRange(None, Date(1970, 2, 2))),
        (True, DateRange(None, Date(1970, 2, 3))),
        (False, DateRange(Date(1970, 2, 1), Date(1970, 2, 2))),
        (False, DateRange(Date(1970, 2, 2), Date(1970, 2, 3))),
        (False, DateRange(Date(1970, 2, 1), Date(1970, 2, 3))),
    ])
    def test_lt_with_start_date(self, expected, other):
        self.assertEquals(expected, DateRange(Date(1970, 2, 2)) < other)

    @parameterized.expand([
        (False, Date(1970, 2, 1)),
        (True, Date(1970, 2, 2)),
        (True, Date(1970, 2, 3)),
        (False, DateRange(Date(1970, 2, 1))),
        (True, DateRange(Date(1970, 2, 2))),
        (True, DateRange(Date(1970, 2, 3))),
        (False, DateRange(None, Date(1970, 2, 1))),
        (False, DateRange(None, Date(1970, 2, 2))),
        (True, DateRange(None, Date(1970, 2, 3))),
        (False, DateRange(Date(1970, 2, 1), Date(1970, 2, 2))),
        (True, DateRange(Date(1970, 2, 2), Date(1970, 2, 3))),
        (False, DateRange(Date(1970, 2, 1), Date(1970, 2, 3))),
    ])
    def test_lt_with_end_date(self, expected, other):
        self.assertEquals(expected, DateRange(None, Date(1970, 2, 2)) < other)

    @parameterized.expand([
        (False, Date(1970, 2, 1)),
        (True, Date(1970, 2, 2)),
        (True, Date(1970, 2, 3)),
        (True, DateRange(Date(1970, 2, 1))),
        (True, DateRange(Date(1970, 2, 2))),
        (True, DateRange(Date(1970, 2, 3))),
        (False, DateRange(None, Date(1970, 2, 1))),
        (True, DateRange(None, Date(1970, 2, 2))),
        (True, DateRange(None, Date(1970, 2, 3))),
        (False, DateRange(Date(1970, 2, 1), Date(1970, 2, 2))),
        (True, DateRange(Date(1970, 2, 2), Date(1970, 2, 3))),
        (False, DateRange(Date(1970, 2, 1), Date(1970, 2, 3))),
    ])
    def test_lt_with_both_dates(self, expected, other):
        self.assertEquals(expected, DateRange(Date(1970, 2, 1), Date(1970, 2, 3)) < other)

    @parameterized.expand([
        (True, DateRange(Date(1970, 2, 2))),
        (False, DateRange(Date(1970, 2, None))),
        (False, DateRange(Date(1970, None, 2))),
        (False, DateRange(Date(None, 2, 2))),
        (False, DateRange(Date(1970, None, None))),
        (False, DateRange(Date(None, 2, None))),
        (False, DateRange(Date(None, None, 2))),
        (False, None),
    ])
    def test_eq(self, expected, other):
        self.assertEquals(expected, DateRange(Date(1970, 2, 2)) == other)

    @parameterized.expand([
        (True, Date(1970, 2, 1)),
        (True, Date(1970, 2, 2)),
        (False, Date(1970, 2, 3)),
        (True, DateRange(Date(1970, 2, 1))),
        (False, DateRange(Date(1970, 2, 2))),
        (False, DateRange(Date(1970, 2, 3))),
        (True, DateRange(None, Date(1970, 2, 1))),
        (True, DateRange(None, Date(1970, 2, 2))),
        (False, DateRange(None, Date(1970, 2, 3))),
        (True, DateRange(Date(1970, 2, 1), Date(1970, 2, 2))),
        (True, DateRange(Date(1970, 2, 2), Date(1970, 2, 3))),
        (True, DateRange(Date(1970, 2, 1), Date(1970, 2, 3))),
    ])
    def test_gt(self, expected, other):
        self.assertEquals(expected, DateRange(Date(1970, 2, 2)) > other)


class NegotiateLocaleTest(TestCase):
    @parameterized.expand([
        ('nl', 'nl', ['nl']),
        ('nl-NL', 'nl', ['nl-NL']),
        ('nl', 'nl-NL', ['nl']),
        ('nl-NL', 'nl-NL', ['nl', 'nl-BE', 'nl-NL']),
        ('nl', 'nl', ['nl', 'en']),
        ('nl', 'nl', ['en', 'nl']),
        ('nl-NL', 'nl-BE', ['nl-NL'])
    ])
    def test(self, expected: Optional[str], preferred_locale: str, available_locales: List[str]):
        self.assertEqual(expected, negotiate_locale(preferred_locale, available_locales))


class NegotiateLocalizedsTest(TestCase):
    class DummyLocalized(Localized):
        def __eq__(self, other):
            return self.locale == other.locale

        def __repr__(self):
            return '%s(%s)' % (self.__class__.__name__, self.locale)

    @parameterized.expand([
        (DummyLocalized('nl'), 'nl', [DummyLocalized('nl')]),
        (DummyLocalized('nl-NL'), 'nl', [DummyLocalized('nl-NL')]),
        (DummyLocalized('nl'), 'nl-NL', [DummyLocalized('nl')]),
        (DummyLocalized('nl'), 'nl', [
         DummyLocalized('nl'), DummyLocalized('en')]),
        (DummyLocalized('nl'), 'nl', [
         DummyLocalized('en'), DummyLocalized('nl')]),
        (None, 'nl', []),
    ])
    def test_with_match_should_return_match(self, expected: Localized, preferred_locale: str, localizeds: List[Localized]):
        self.assertEquals(expected, negotiate_localizeds(
            preferred_locale, localizeds))

    def test_without_match_should_return_default(self):
        preferred_locale = 'de'
        localizeds = [self.DummyLocalized('nl'), self.DummyLocalized(
            'en'), self.DummyLocalized('uk')]
        self.assertEquals(self.DummyLocalized('nl'), negotiate_localizeds(
            preferred_locale, localizeds))


_FORMAT_DATE_TEST_PARAMETERS = [
    # Dates that cannot be formatted.
    ('unknown date', Date()),
    ('unknown date', Date(None, None, 1)),
    # Single dates.
    ('January', Date(None, 1, None)),
    ('around January', Date(None, 1, None, fuzzy=True)),
    ('1970', Date(1970, None, None)),
    ('around 1970', Date(1970, None, None, fuzzy=True)),
    ('January, 1970', Date(1970, 1, None)),
    ('around January, 1970', Date(1970, 1, None, fuzzy=True)),
    ('January 1, 1970', Date(1970, 1, 1)),
    ('around January 1, 1970', Date(1970, 1, 1, fuzzy=True)),
    ('January 1', Date(None, 1, 1)),
    ('around January 1', Date(None, 1, 1, fuzzy=True)),
]


class FormatDateTest(TestCase):
    @parameterized.expand(_FORMAT_DATE_TEST_PARAMETERS)
    def test(self, expected: str, datey: Datey):
        locale = 'en'
        with Translations(gettext.NullTranslations()):
            self.assertEquals(expected, format_datey(datey, locale))


_FORMAT_DATE_RANGE_TEST_PARAMETERS = [
    ('from January 1, 1970 until December 31, 1999', DateRange(Date(1970, 1, 1), Date(1999, 12, 31))),
    ('from January 1, 1970 until sometime before December 31, 1999', DateRange(Date(1970, 1, 1), Date(1999, 12, 31), end_is_boundary=True)),
    ('from January 1, 1970 until around December 31, 1999', DateRange(Date(1970, 1, 1), Date(1999, 12, 31, fuzzy=True))),
    ('from January 1, 1970 until sometime before around December 31, 1999', DateRange(Date(1970, 1, 1), Date(1999, 12, 31, fuzzy=True), end_is_boundary=True)),
    ('from sometime after January 1, 1970 until December 31, 1999', DateRange(Date(1970, 1, 1), Date(1999, 12, 31), start_is_boundary=True)),
    ('sometime between January 1, 1970 and December 31, 1999', DateRange(Date(1970, 1, 1), Date(1999, 12, 31), start_is_boundary=True, end_is_boundary=True)),
    ('from sometime after January 1, 1970 until around December 31, 1999', DateRange(Date(1970, 1, 1), Date(1999, 12, 31, fuzzy=True), start_is_boundary=True)),
    ('sometime between January 1, 1970 and around December 31, 1999', DateRange(Date(1970, 1, 1), Date(1999, 12, 31, fuzzy=True), start_is_boundary=True, end_is_boundary=True)),
    ('from around January 1, 1970 until December 31, 1999', DateRange(Date(1970, 1, 1, fuzzy=True), Date(1999, 12, 31))),
    ('from around January 1, 1970 until sometime before December 31, 1999', DateRange(Date(1970, 1, 1, fuzzy=True), Date(1999, 12, 31), end_is_boundary=True)),
    ('from around January 1, 1970 until around December 31, 1999', DateRange(Date(1970, 1, 1, fuzzy=True), Date(1999, 12, 31, fuzzy=True))),
    ('from around January 1, 1970 until sometime before around December 31, 1999', DateRange(Date(1970, 1, 1, fuzzy=True), Date(1999, 12, 31, fuzzy=True), end_is_boundary=True)),
    ('from sometime after around January 1, 1970 until December 31, 1999', DateRange(Date(1970, 1, 1, fuzzy=True), Date(1999, 12, 31), start_is_boundary=True)),
    ('sometime between around January 1, 1970 and December 31, 1999', DateRange(Date(1970, 1, 1, fuzzy=True), Date(1999, 12, 31), start_is_boundary=True, end_is_boundary=True)),
    ('from sometime after around January 1, 1970 until around December 31, 1999', DateRange(Date(1970, 1, 1, fuzzy=True), Date(1999, 12, 31, fuzzy=True), start_is_boundary=True)),
    ('sometime between around January 1, 1970 and around December 31, 1999', DateRange(Date(1970, 1, 1, fuzzy=True), Date(1999, 12, 31, fuzzy=True), start_is_boundary=True, end_is_boundary=True)),
    ('from January 1, 1970', DateRange(Date(1970, 1, 1))),
    ('sometime after January 1, 1970', DateRange(Date(1970, 1, 1), start_is_boundary=True)),
    ('from around January 1, 1970', DateRange(Date(1970, 1, 1, fuzzy=True))),
    ('sometime after around January 1, 1970', DateRange(Date(1970, 1, 1, fuzzy=True), start_is_boundary=True)),
    ('until December 31, 1999', DateRange(None, Date(1999, 12, 31))),
    ('sometime before December 31, 1999', DateRange(None, Date(1999, 12, 31), end_is_boundary=True)),
    ('until around December 31, 1999', DateRange(None, Date(1999, 12, 31, fuzzy=True))),
    ('sometime before around December 31, 1999', DateRange(None, Date(1999, 12, 31, fuzzy=True), end_is_boundary=True)),
]


class FormatDateRangeTest(TestCase):
    @parameterized.expand(_FORMAT_DATE_RANGE_TEST_PARAMETERS)
    def test(self, expected: str, datey: Datey):
        locale = 'en'
        with Translations(gettext.NullTranslations()):
            self.assertEquals(expected, format_datey(datey, locale))


class FormatDateyTest(TestCase):
    @parameterized.expand(_FORMAT_DATE_TEST_PARAMETERS + _FORMAT_DATE_RANGE_TEST_PARAMETERS)
    def test(self, expected: str, datey: Datey):
        locale = 'en'
        with Translations(gettext.NullTranslations()):
            self.assertEquals(expected, format_datey(datey, locale))


class OpenTranslationsTest(TestCase):
    def test(self) -> None:
        locale = 'nl-NL'
        locale_path_name = 'nl_NL'
        with TemporaryDirectory() as assets_directory_path:
            lc_messages_directory_path = path.join(assets_directory_path, 'locale', locale_path_name, 'LC_MESSAGES')
            makedirs(lc_messages_directory_path)
            po = """
# Dutch translations for PROJECT.
# Copyright (C) 2019 ORGANIZATION
# This file is distributed under the same license as the PROJECT project.
# FIRST AUTHOR <EMAIL@ADDRESS>, 2019.
#
msgid ""
msgstr ""
"Project-Id-Version: PROJECT VERSION\n"
"Report-Msgid-Bugs-To: EMAIL@ADDRESS\n"
"POT-Creation-Date: 2020-11-18 23:28+0000\n"
"PO-Revision-Date: 2019-10-05 11:38+0100\n"
"Last-Translator: \n"
"Language: nl\n"
"Language-Team: nl <LL@li.org>\n"
"Plural-Forms: nplurals=2; plural=(n != 1)\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=utf-8\n"
"Content-Transfer-Encoding: 8bit\n"
"Generated-By: Babel 2.7.0\n"

#: betty/ancestry.py:457
msgid "Subject"
msgstr "Onderwerp"
"""
            with open(path.join(lc_messages_directory_path, 'betty.po'), 'w') as f:
                f.write(po)
            self.assertIsInstance(open_translations(locale, assets_directory_path), gettext.NullTranslations)



from betty.graph import tsort, CyclicGraphError
from betty.tests import TestCase


class TSortTest(TestCase):
    def test_with_empty_graph(self):
        graph = {}
        self.assertCountEqual([], tsort(graph))

    def test_with_isolated_vertices(self):
        graph = {
            1: set(),
            2: set(),
        }
        # Without edges we cannot assert the order.
        self.assertCountEqual([1, 2], tsort(graph))

    def test_with_edges(self):
        graph = {
            1: {2},
        }
        self.assertEquals([1, 2], tsort(graph))

    def test_with_multiple_chained_edges(self):
        graph = {
            2: {3},
            1: {2},
        }
        self.assertEquals([1, 2, 3], tsort(graph))

    def test_with_multiple_indegrees(self):
        graph = {
            1: {3},
            2: {3},
        }
        vertices = tsort(graph)
        self.assertEquals(3, len(vertices))
        self.assertIn(1, vertices)
        self.assertIn(2, vertices)
        self.assertEquals(3, vertices[2])

    def test_with_multiple_outdegrees(self):
        graph = {
            1: {2, 3},
        }
        vertices = tsort(graph)
        self.assertEquals(3, len(vertices))
        self.assertEquals(1, vertices[0])
        self.assertIn(2, vertices)
        self.assertIn(3, vertices)

    def test_with_cyclic_edges(self):
        graph = {
            1: {2},
            2: {1},
        }
        with self.assertRaises(CyclicGraphError):
            tsort(graph)



import json
from collections import OrderedDict
from contextlib import contextmanager
from os.path import join
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Any, Dict

import yaml
from parameterized import parameterized
from voluptuous import Schema, Required

from betty.config import from_file, Configuration, ConfigurationValueError, LocaleConfiguration, PluginsConfiguration
from betty.plugin import Plugin, NO_CONFIGURATION
from betty.site import Site
from betty.tests import TestCase


class NonConfigurablePlugin(Plugin):
    pass  # pragma: no cover


class ConfigurablePlugin(Plugin):
    configuration_schema: Schema = Schema({
        Required('check'): lambda x: x
    })

    def __init__(self, check):
        self.check = check

    @classmethod
    def for_site(cls, site: Site, configuration: Any = NO_CONFIGURATION):
        return cls(configuration['check'])


class PluginsConfigurationTest(TestCase):
    def test_init_from_dict_with_valid_configuration_should_set(self):
        plugin_configuration = {
            'check': 1337,
        }
        plugins_configuration_dict = {
            ConfigurablePlugin: plugin_configuration,
        }
        sut = PluginsConfiguration(plugins_configuration_dict)
        self.assertEqual(plugin_configuration, sut[ConfigurablePlugin])

    def test_init_from_dict_with_invalid_configuration_should_raise_configuration_error(self):
        plugin_configuration = 1337
        plugins_configuration_dict = {
            ConfigurablePlugin: plugin_configuration,
        }
        with self.assertRaises(ConfigurationValueError):
            PluginsConfiguration(plugins_configuration_dict)

    def test_setitem_and_getitem_with_valid_configuration_should_set_and_return(self):
        plugin_configuration = {
            'check': 1337,
        }
        sut = PluginsConfiguration()
        sut[ConfigurablePlugin] = plugin_configuration
        self.assertEqual(plugin_configuration, sut[ConfigurablePlugin])

    def test_setitem_with_invalid_configuration_should_raise_configuration_error(self):
        plugin_configuration = 1337
        sut = PluginsConfiguration()
        with self.assertRaises(ConfigurationValueError):
            sut[ConfigurablePlugin] = plugin_configuration

    def test_contains(self):
        sut = PluginsConfiguration()
        sut[NonConfigurablePlugin] = None
        self.assertIn(NonConfigurablePlugin, sut)

    def test_iter(self):
        sut = PluginsConfiguration()
        sut[NonConfigurablePlugin] = None
        self.assertSequenceEqual([(NonConfigurablePlugin, None)], list(sut))

    def test_len(self):
        sut = PluginsConfiguration()
        sut[NonConfigurablePlugin] = None
        self.assertEqual(1, len(sut))


class LocaleConfigurationTest(TestCase):
    def test_locale(self):
        locale = 'nl-NL'
        sut = LocaleConfiguration(locale)
        self.assertEquals(locale, sut.locale)

    def test_alias_implicit(self):
        locale = 'nl-NL'
        sut = LocaleConfiguration(locale)
        self.assertEquals(locale, sut.alias)

    def test_alias_explicit(self):
        locale = 'nl-NL'
        alias = 'nl'
        sut = LocaleConfiguration(locale, alias)
        self.assertEquals(alias, sut.alias)

    @parameterized.expand([
        (False, LocaleConfiguration('nl', 'NL'), 'not a locale configuration'),
        (False, LocaleConfiguration('nl', 'NL'), 999),
        (False, LocaleConfiguration('nl', 'NL'), object()),
    ])
    def test_eq(self, expected, sut, other):
        self.assertEquals(expected, sut == other)


class ConfigurationTest(TestCase):
    def test_output_directory_path(self):
        output_directory_path = '/tmp/betty'
        sut = Configuration(output_directory_path, 'https://example.com')
        self.assertEquals(output_directory_path, sut.output_directory_path)

    def test_www_directory_path_with_absolute_path(self):
        output_directory_path = '/tmp/betty'
        sut = Configuration(output_directory_path, 'https://example.com')
        expected = join(output_directory_path, 'www')
        self.assertEquals(expected, sut.www_directory_path)

    def test_assets_directory_path_without_path(self):
        sut = Configuration('/tmp/betty', 'https://example.com')
        self.assertIsNone(sut.assets_directory_path)

    def test_assets_directory_path_with_path(self):
        sut = Configuration('/tmp/betty', 'https://example.com')
        assets_directory_path = '/tmp/betty-assets'
        sut.assets_directory_path = assets_directory_path
        self.assertEquals(assets_directory_path,
                          sut.assets_directory_path)

    def test_root_path(self):
        sut = Configuration('/tmp/betty', 'https://example.com')
        configured_root_path = '/betty'
        expected_root_path = '/betty/'
        sut.root_path = configured_root_path
        self.assertEquals(expected_root_path, sut.root_path)

    def test_clean_urls(self):
        sut = Configuration('/tmp/betty', 'https://example.com')
        clean_urls = True
        sut.clean_urls = clean_urls
        self.assertEquals(clean_urls, sut.clean_urls)

    def test_content_negotiation(self):
        sut = Configuration('/tmp/betty', 'https://example.com')
        content_negotiation = True
        sut.content_negotiation = content_negotiation
        self.assertEquals(content_negotiation, sut.content_negotiation)

    def test_clean_urls_implied_by_content_negotiation(self):
        sut = Configuration('/tmp/betty', 'https://example.com')
        sut.content_negotiation = True
        self.assertTrue(sut.clean_urls)

    def test_author_without_author(self):
        sut = Configuration('/tmp/betty', 'https://example.com')
        self.assertIsNone(sut.author)

    def test_author_with_author(self):
        sut = Configuration('/tmp/betty', 'https://example.com')
        author = 'Bart'
        sut.author = author
        self.assertEquals(author, sut.author)


class FromTest(TestCase):
    @contextmanager
    def _build_minimal_config(self) -> Dict:
        output_directory = TemporaryDirectory()
        try:
            yield {
                'output': output_directory.name,
                'base_url': 'https://example.com',
            }
        finally:
            output_directory.cleanup()

    def _writes(self, config: str, extension: str):
        f = NamedTemporaryFile(mode='r+', suffix='.' + extension)
        f.write(config)
        f.seek(0)
        return f

    def _write(self, config_dict: Dict[str, Any]):
        return self._writes(json.dumps(config_dict), 'json')

    @parameterized.expand([
        ('json', json.dumps),
        ('yaml', yaml.safe_dump),
        ('yml', yaml.safe_dump),
    ])
    def test_from_file_should_parse_minimal(self, extension, dumper):
        with self._build_minimal_config() as config_dict:
            with self._writes(dumper(config_dict), extension) as f:
                configuration = from_file(f)
            self.assertEquals(config_dict['output'], configuration.output_directory_path)
            self.assertEquals(config_dict['base_url'], configuration.base_url)
            self.assertEquals('Betty', configuration.title)
            self.assertIsNone(configuration.author)
            self.assertEquals('production', configuration.mode)
            self.assertEquals('/', configuration.root_path)
            self.assertFalse(configuration.clean_urls)
            self.assertFalse(configuration.content_negotiation)

    def test_from_file_should_parse_title(self):
        title = 'My first Betty site'
        with self._build_minimal_config() as config_dict:
            config_dict['title'] = title
            with self._write(config_dict) as f:
                configuration = from_file(f)
            self.assertEquals(title, configuration.title)

    def test_from_file_should_parse_author(self):
        author = 'Bart'
        with self._build_minimal_config() as config_dict:
            config_dict['author'] = author
            with self._write(config_dict) as f:
                configuration = from_file(f)
            self.assertEquals(author, configuration.author)

    def test_from_file_should_parse_locale_locale(self):
        locale = 'nl-NL'
        locale_config = {
            'locale': locale,
        }
        with self._build_minimal_config() as config_dict:
            config_dict['locales'] = [locale_config]
            with self._write(config_dict) as f:
                configuration = from_file(f)
            self.assertDictEqual(OrderedDict({
                locale: LocaleConfiguration(locale),
            }), configuration.locales)

    def test_from_file_should_parse_locale_alias(self):
        locale = 'nl-NL'
        alias = 'nl'
        locale_config = {
            'locale': locale,
            'alias': alias,
        }
        with self._build_minimal_config() as config_dict:
            config_dict['locales'] = [locale_config]
            with self._write(config_dict) as f:
                configuration = from_file(f)
            self.assertDictEqual(OrderedDict({
                locale: LocaleConfiguration(locale, alias),
            }), configuration.locales)

    def test_from_file_should_root_path(self):
        configured_root_path = '/betty'
        expected_root_path = '/betty/'
        with self._build_minimal_config() as config_dict:
            config_dict['root_path'] = configured_root_path
            with self._write(config_dict) as f:
                configuration = from_file(f)
            self.assertEquals(expected_root_path, configuration.root_path)

    def test_from_file_should_clean_urls(self):
        clean_urls = True
        with self._build_minimal_config() as config_dict:
            config_dict['clean_urls'] = clean_urls
            with self._write(config_dict) as f:
                configuration = from_file(f)
            self.assertEquals(clean_urls, configuration.clean_urls)

    def test_from_file_should_content_negotiation(self):
        content_negotiation = True
        with self._build_minimal_config() as config_dict:
            config_dict['content_negotiation'] = content_negotiation
            with self._write(config_dict) as f:
                configuration = from_file(f)
            self.assertEquals(content_negotiation, configuration.content_negotiation)

    @parameterized.expand([
        ('production',),
        ('development',),
    ])
    def test_from_file_should_parse_mode(self, mode: str):
        with self._build_minimal_config() as config_dict:
            config_dict['mode'] = mode
            with self._write(config_dict) as f:
                configuration = from_file(f)
            self.assertEquals(mode, configuration.mode)

    def test_from_file_should_parse_assets_directory_path(self):
        with TemporaryDirectory() as assets_directory_path:
            with self._build_minimal_config() as config_dict:
                config_dict['assets_directory_path'] = assets_directory_path
                with self._write(config_dict) as f:
                    configuration = from_file(f)
                self.assertEquals(assets_directory_path, configuration.assets_directory_path)

    def test_from_file_should_parse_one_plugin_with_configuration(self):
        with self._build_minimal_config() as config_dict:
            plugin_configuration = {
                'check': 1337,
            }
            config_dict['plugins'] = {
                ConfigurablePlugin.name(): plugin_configuration,
            }
            with self._write(config_dict) as f:
                configuration = from_file(f)
            expected = {
                ConfigurablePlugin: plugin_configuration,
            }
            self.assertEquals(expected, dict(configuration.plugins))

    def test_from_file_should_parse_one_plugin_without_configuration(self):
        with self._build_minimal_config() as config_dict:
            config_dict['plugins'] = {
                NonConfigurablePlugin.name(): None,
            }
            with self._write(config_dict) as f:
                configuration = from_file(f)
            expected = {
                NonConfigurablePlugin: None,
            }
            self.assertEquals(expected, dict(configuration.plugins))

    def test_from_file_should_error_unknown_format(self):
        with self._writes('', 'abc') as f:
            with self.assertRaises(ConfigurationValueError):
                from_file(f)

    def test_from_file_should_error_if_invalid_json(self):
        with self._writes('', 'json') as f:
            with self.assertRaises(ConfigurationValueError):
                from_file(f)

    def test_from_file_should_error_if_invalid_yaml(self):
        with self._writes('"foo', 'yaml') as f:
            with self.assertRaises(ConfigurationValueError):
                from_file(f)

    def test_from_file_should_error_if_invalid_config(self):
        config_dict = {}
        with self._write(config_dict) as f:
            with self.assertRaises(ConfigurationValueError):
                from_file(f)



from gettext import NullTranslations
from tempfile import TemporaryFile, NamedTemporaryFile
from typing import Any
from unittest.mock import Mock

from geopy import Point
from parameterized import parameterized

from betty.ancestry import EventHandlingSetList, Person, Event, Place, File, Note, Presence, PlaceName, PersonName, \
    IdentifiableEvent, Subject, Birth, Enclosure, Identifiable, Described, Dated, HasPrivacy, HasMediaType, Link, \
    HasLinks, HasNotes, HasFiles, Source, Citation, HasCitations, IdentifiableCitation, IdentifiableSource, \
    PresenceRole, Attendee, Beneficiary, Witness, EventType, UnknownEventType, LifeEventType, PreBirthEventType, \
    PostDeathEventType, Baptism, Adoption, Death, Funeral, FinalDispositionEventType, Cremation, Burial, Will, \
    Engagement, Marriage, MarriageAnnouncement, Divorce, DivorceAnnouncement, Residence, Immigration, Emigration, \
    Correspondence, Occupation, Retirement, Confirmation, Missing, EVENT_TYPE_TYPES, RESOURCE_TYPES, Resource
from betty.locale import Date, Translations
from betty.media_type import MediaType
from betty.tests import TestCase


class EventHandlingSetListTest(TestCase):
    def test_prepend(self) -> None:
        added = []
        removal_handler = Mock()
        sut = EventHandlingSetList(lambda value: added.append(value), removal_handler)
        sut.prepend(3)
        sut.prepend(2)
        sut.prepend(1)
        # Prepend an already prepended value again, and assert that it was ignored.
        sut.prepend(1)
        self.assertSequenceEqual([1, 2, 3], sut)
        self.assertSequenceEqual([3, 2, 1], added)
        removal_handler.assert_not_called()

    def test_append(self) -> None:
        added = []
        removal_handler = Mock()
        sut = EventHandlingSetList(lambda value: added.append(value), removal_handler)
        sut.append(3)
        sut.append(2)
        sut.append(1)
        # Append an already appended value again, and assert that it was ignored.
        sut.append(1)
        self.assertSequenceEqual([3, 2, 1], sut)
        self.assertSequenceEqual([3, 2, 1], added)
        removal_handler.assert_not_called()

    def test_remove(self) -> None:
        added = []
        removed = []
        sut = EventHandlingSetList(lambda value: added.append(value), lambda value: removed.append(value))
        sut.append(1, 2, 3, 4)
        sut.remove(4, 2)
        self.assertSequenceEqual([1, 3], sut)
        self.assertSequenceEqual([1, 2, 3, 4], added)
        self.assertSequenceEqual([4, 2], removed)

    def test_replace(self) -> None:
        added = []
        removed = []
        sut = EventHandlingSetList(lambda value: added.append(value), lambda value: removed.append(value))
        sut.append(1, 2, 3)
        sut.replace(4, 5, 6)
        self.assertSequenceEqual([4, 5, 6], sut)
        self.assertSequenceEqual([1, 2, 3, 4, 5, 6], added)
        self.assertSequenceEqual([1, 2, 3], removed)

    def test_clear(self) -> None:
        added = []
        removed = []
        sut = EventHandlingSetList(lambda value: added.append(value), lambda value: removed.append(value))
        sut.append(1, 2, 3)
        sut.clear()
        self.assertSequenceEqual([], sut)
        self.assertSequenceEqual([1, 2, 3], added)
        self.assertSequenceEqual([1, 2, 3], removed)

    def test_list(self) -> None:
        sut = EventHandlingSetList(lambda _: None, lambda _: None)
        sut.append(1, 2, 3)
        self.assertEqual([1, 2, 3], sut.list)

    def test_len(self) -> None:
        sut = EventHandlingSetList(lambda _: None, lambda _: None)
        sut.append(1, 2, 3)
        self.assertEqual(3, len(sut))

    def test_iter(self) -> None:
        sut = EventHandlingSetList(lambda _: None, lambda _: None)
        sut.append(1, 2, 3)
        # list() gets all items through __iter__ and stores them in the same order.
        self.assertSequenceEqual([1, 2, 3], list(sut))

    def test_getitem(self) -> None:
        sut = EventHandlingSetList(lambda _: None, lambda _: None)
        sut.append(1, 2, 3)
        self.assertEqual(1, sut[0])
        self.assertEqual(2, sut[1])
        self.assertEqual(3, sut[2])
        with self.assertRaises(IndexError):
            sut[3]

    def test_set_like_functionality(self) -> None:
        sut = EventHandlingSetList(lambda _: None, lambda _: None)
        # Ensure duplicates are skipped.
        sut.append(1, 2, 3, 1, 2, 3, 1, 2, 3)
        # Ensure skipped duplicates do not affect further new values.
        sut.append(1, 2, 3, 4, 5, 6, 7, 8, 9)
        self.assertSequenceEqual([1, 2, 3, 4, 5, 6, 7, 8, 9], sut)


class HasPrivacyTest(TestCase):
    def test_date(self) -> None:
        sut = HasPrivacy()
        self.assertIsNone(sut.private)


class DatedTest(TestCase):
    def test_date(self) -> None:
        sut = Dated()
        self.assertIsNone(sut.date)


class NoteTest(TestCase):
    def test_text(self) -> None:
        text = 'Betty wrote this.'
        sut = Note(text)
        self.assertEquals(text, sut.text)


class HasNotesTest(TestCase):
    def test_notes(self) -> None:
        sut = HasNotes()
        self.assertEquals([], sut.notes)


class IdentifiableTest(TestCase):
    def test_id(self) -> None:
        identifiable_id = '000000001'
        sut = Identifiable(identifiable_id)
        self.assertEquals(identifiable_id, sut.id)


class DescribedTest(TestCase):
    def test_description(self) -> None:
        sut = Described()
        self.assertIsNone(sut.description)


class HasMediaTypeTest(TestCase):
    def test_media_type(self) -> None:
        sut = HasMediaType()
        self.assertIsNone(sut.media_type)


class LinkTest(TestCase):
    def test_url(self) -> None:
        url = 'https://example.com'
        sut = Link(url)
        self.assertEquals(url, sut.url)

    def test_media_type(self) -> None:
        url = 'https://example.com'
        sut = Link(url)
        self.assertIsNone(sut.media_type)

    def test_locale(self) -> None:
        url = 'https://example.com'
        sut = Link(url)
        self.assertIsNone(sut.locale)

    def test_description(self) -> None:
        url = 'https://example.com'
        sut = Link(url)
        self.assertIsNone(sut.description)

    def test_relationship(self) -> None:
        url = 'https://example.com'
        sut = Link(url)
        self.assertIsNone(sut.relationship)

    def test_label(self) -> None:
        url = 'https://example.com'
        sut = Link(url)
        self.assertIsNone(sut.label)


class HasLinksTest(TestCase):
    def test_links(self) -> None:
        sut = HasLinks()
        self.assertEquals(set(), sut.links)


class FileTest(TestCase):
    def test_resource_type_name(self) -> None:
        self.assertIsInstance(File.resource_type_name(), str)
        self.assertNotEqual('', File.resource_type_name())

    def test_id(self) -> None:
        file_id = 'BETTY01'
        file_path = '/tmp/betty'
        sut = File(file_id, file_path)
        self.assertEquals(file_id, sut.id)

    def test_private(self) -> None:
        file_id = 'BETTY01'
        file_path = '/tmp/betty'
        sut = File(file_id, file_path)
        self.assertIsNone(sut.private)
        private = True
        sut.private = private
        self.assertEquals(private, sut.private)

    def test_media_type(self) -> None:
        file_id = 'BETTY01'
        file_path = '/tmp/betty'
        sut = File(file_id, file_path)
        self.assertIsNone(sut.media_type)
        media_type = MediaType('text/plain')
        sut.media_type = media_type
        self.assertEquals(media_type, sut.media_type)

    def test_path(self) -> None:
        with TemporaryFile() as f:
            file_id = 'BETTY01'
            sut = File(file_id, f.name)
            self.assertEquals(f.name, sut.path)

    def test_extension_with_extension(self) -> None:
        extension = 'betty'
        with NamedTemporaryFile(suffix='.%s' % extension) as f:
            file_id = 'BETTY01'
            sut = File(file_id, f.name)
            self.assertEquals(extension, sut.extension)

    def test_extension_without_extension(self) -> None:
        with NamedTemporaryFile() as f:
            file_id = 'BETTY01'
            sut = File(file_id, f.name)
            self.assertIsNone(sut.extension)

    def test_description(self) -> None:
        file_id = 'BETTY01'
        file_path = '/tmp/betty'
        sut = File(file_id, file_path)
        self.assertIsNone(sut.description)
        description = 'Hi, my name is Betty!'
        sut.description = description
        self.assertEquals(description, sut.description)

    def test_notes(self) -> None:
        file_id = 'BETTY01'
        file_path = '/tmp/betty'
        sut = File(file_id, file_path)
        self.assertCountEqual([], sut.notes)
        notes = [Mock(Note), Mock(Note)]
        sut.notes = notes
        self.assertCountEqual(notes, sut.notes)

    def test_resources(self) -> None:
        file_id = 'BETTY01'
        file_path = '/tmp/betty'
        sut = File(file_id, file_path)
        self.assertCountEqual([], sut.resources)
        resources = [Mock(HasFiles), Mock(HasFiles)]
        sut.resources = resources
        self.assertCountEqual(resources, sut.resources)

    def test_sources(self) -> None:
        file_id = 'BETTY01'
        file_path = '/tmp/betty'
        sut = File(file_id, file_path)
        self.assertCountEqual([], sut.resources)
        source = Mock(Source)
        citation_source = Mock(Source)
        citation = Citation(citation_source)
        resources = [citation, source, Mock(HasFiles)]
        sut.resources = resources
        self.assertCountEqual([citation_source, source], sut.sources)

    def test_citations(self) -> None:
        file_id = 'BETTY01'
        file_path = '/tmp/betty'
        sut = File(file_id, file_path)
        self.assertCountEqual([], sut.resources)
        citation = Mock(Citation)
        resources = [citation, Mock(HasFiles)]
        sut.resources = resources
        self.assertCountEqual([citation], sut.citations)

    def test_name(self) -> None:
        file_id = 'BETTY01'
        file_path = '/tmp/betty.png'
        sut = File(file_id, file_path)
        self.assertEquals('betty.png', sut.name)

    def test_basename(self) -> None:
        file_id = 'BETTY01'
        file_path = '/tmp/betty.png'
        sut = File(file_id, file_path)
        self.assertEquals('/tmp/betty', sut.basename)

    def test_extension(self) -> None:
        file_id = 'BETTY01'
        file_path = '/tmp/betty.png'
        sut = File(file_id, file_path)
        self.assertEquals('png', sut.extension)


class HasFilesTest(TestCase):
    def test_files(self) -> None:
        sut = HasFiles()
        self.assertCountEqual([], sut.files)
        files = [Mock(File), Mock(File)]
        sut.files = files
        self.assertCountEqual(files, sut.files)


class SourceTest(TestCase):
    def test_resource_type_name(self) -> None:
        self.assertIsInstance(Source.resource_type_name(), str)
        self.assertNotEqual('', Source.resource_type_name())

    def test_name(self) -> None:
        name = 'The Source'
        sut = Source(name)
        self.assertEquals(name, sut.name)

    def test_contained_by(self) -> None:
        contained_by_source = Source()
        sut = Source()
        self.assertIsNone(sut.contained_by)
        sut.contained_by = contained_by_source
        self.assertEquals(contained_by_source, sut.contained_by)

    def test_contains(self) -> None:
        contains_source = Source()
        sut = Source()
        self.assertCountEqual([], sut.contains)
        sut.contains = [contains_source]
        self.assertCountEqual([contains_source], sut.contains)

    def test_citations(self) -> None:
        sut = Source()
        self.assertCountEqual([], sut.citations)

    def test_author(self) -> None:
        sut = Source()
        self.assertIsNone(sut.author)
        author = 'Me'
        sut.author = author
        self.assertEquals(author, sut.author)

    def test_publisher(self) -> None:
        sut = Source()
        self.assertIsNone(sut.publisher)
        publisher = 'Me'
        sut.publisher = publisher
        self.assertEquals(publisher, sut.publisher)

    def test_date(self) -> None:
        sut = Source()
        self.assertIsNone(sut.date)

    def test_files(self) -> None:
        sut = Source()
        self.assertCountEqual([], sut.files)

    def test_links(self) -> None:
        sut = Source()
        self.assertCountEqual([], sut.links)

    def test_private(self) -> None:
        sut = Source()
        self.assertIsNone(sut.private)
        private = True
        sut.private = private
        self.assertEquals(private, sut.private)


class IdentifiableSourceTest(TestCase):
    def test_id(self) -> None:
        source_id = 'C1'
        sut = IdentifiableSource(source_id)
        self.assertEquals(source_id, sut.id)


class CitationTest(TestCase):
    def test_resource_type_name(self) -> None:
        self.assertIsInstance(Citation.resource_type_name(), str)
        self.assertNotEqual('', Citation.resource_type_name())

    def test_facts(self) -> None:
        fact = Mock(HasCitations)
        sut = Citation(Mock(Source))
        self.assertCountEqual([], sut.facts)
        sut.facts = [fact]
        self.assertCountEqual([fact], sut.facts)

    def test_source(self) -> None:
        source = Mock(Source)
        sut = Citation(source)
        self.assertEquals(source, sut.source)

    def test_location(self) -> None:
        sut = Citation(Mock(Source))
        self.assertIsNone(sut.location)
        location = 'Somewhere'
        sut.location = location
        self.assertEquals(location, sut.location)

    def test_date(self) -> None:
        sut = Citation(Mock(Source))
        self.assertIsNone(sut.date)

    def test_files(self) -> None:
        sut = Citation(Mock(Source))
        self.assertCountEqual([], sut.files)

    def test_private(self) -> None:
        sut = Citation(Mock(Source))
        self.assertIsNone(sut.private)
        private = True
        sut.private = private
        self.assertEquals(private, sut.private)


class IdentifiableCitationTest(TestCase):
    def test_id(self) -> None:
        citation_id = 'C1'
        sut = IdentifiableCitation(citation_id, Mock(Source))
        self.assertEquals(citation_id, sut.id)


class HasCitationsTest(TestCase):
    def test_citations(self) -> None:
        sut = HasCitations()
        self.assertCountEqual([], sut.citations)
        citation = Mock(Citation)
        sut.citations = [citation]
        self.assertCountEqual([citation], sut.citations)


class PlaceNameTest(TestCase):
    @parameterized.expand([
        (True, PlaceName('Ikke'), PlaceName('Ikke')),
        (True, PlaceName('Ikke', 'nl-NL'), PlaceName('Ikke', 'nl-NL')),
        (False, PlaceName('Ikke', 'nl-NL'), PlaceName('Ikke', 'nl-BE')),
        (False, PlaceName('Ikke', 'nl-NL'), PlaceName('Ik', 'nl-NL')),
        (False, PlaceName('Ikke'), PlaceName('Ik')),
        (False, PlaceName('Ikke'), None),
        (False, PlaceName('Ikke'), 'not-a-place-name'),
    ])
    def test_eq(self, expected, a, b) -> None:
        self.assertEquals(expected, a == b)

    def test_str(self) -> None:
        name = 'Ikke'
        sut = PlaceName(name)
        self.assertEquals(name, str(sut))

    def test_name(self) -> None:
        name = 'Ikke'
        sut = PlaceName(name)
        self.assertEquals(name, sut.name)

    def test_locale(self) -> None:
        locale = 'nl-NL'
        sut = PlaceName('Ikke', locale=locale)
        self.assertEquals(locale, sut.locale)

    def test_date(self) -> None:
        date = Date()
        sut = PlaceName('Ikke', date=date)
        self.assertEquals(date, sut.date)


class EnclosureTest(TestCase):
    def test_encloses(self) -> None:
        encloses = Mock(Place)
        enclosed_by = Mock(Place)
        sut = Enclosure(encloses, enclosed_by)
        self.assertEquals(encloses, sut.encloses)

    def test_enclosed_by(self) -> None:
        encloses = Mock(Place)
        enclosed_by = Mock(Place)
        sut = Enclosure(encloses, enclosed_by)
        self.assertEquals(enclosed_by, sut.enclosed_by)

    def test_date(self) -> None:
        encloses = Mock(Place)
        enclosed_by = Mock(Place)
        sut = Enclosure(encloses, enclosed_by)
        date = Date()
        self.assertIsNone(sut.date)
        sut.date = date
        self.assertEquals(date, sut.date)

    def test_citations(self) -> None:
        encloses = Mock(Place)
        enclosed_by = Mock(Place)
        sut = Enclosure(encloses, enclosed_by)
        citation = Mock(Citation)
        self.assertIsNone(sut.date)
        sut.citations = [citation]
        self.assertCountEqual([citation], sut.citations)


class PlaceTest(TestCase):
    def test_resource_type_name(self) -> None:
        self.assertIsInstance(Citation.resource_type_name(), str)
        self.assertNotEqual('', Citation.resource_type_name())

    def test_events(self) -> None:
        sut = Place('P1', [PlaceName('The Place')])
        event = IdentifiableEvent('1', Birth())
        sut.events.append(event)
        self.assertIn(event, sut.events)
        self.assertEquals(sut, event.place)
        sut.events.remove(event)
        self.assertCountEqual([], sut.events)
        self.assertEquals(None, event.place)

    def test_enclosed_by(self) -> None:
        sut = Place('P1', [PlaceName('The Place')])
        self.assertCountEqual([], sut.enclosed_by)
        enclosing_place = Place('P2', [PlaceName('The Other Place')])
        enclosure = Enclosure(sut, enclosing_place)
        self.assertIn(enclosure, sut.enclosed_by)
        self.assertEquals(sut, enclosure.encloses)
        sut.enclosed_by.remove(enclosure)
        self.assertCountEqual([], sut.enclosed_by)
        self.assertIsNone(enclosure.encloses)

    def test_encloses(self) -> None:
        sut = Place('P1', [PlaceName('The Place')])
        self.assertCountEqual([], sut.encloses)
        enclosed_place = Place('P2', [PlaceName('The Other Place')])
        enclosure = Enclosure(enclosed_place, sut)
        self.assertIn(enclosure, sut.encloses)
        self.assertEquals(sut, enclosure.enclosed_by)
        sut.encloses.remove(enclosure)
        self.assertCountEqual([], sut.encloses)
        self.assertIsNone(enclosure.enclosed_by)

    def test_id(self) -> None:
        place_id = 'C1'
        sut = Place(place_id, [PlaceName('one')])
        self.assertEquals(place_id, sut.id)

    def test_links(self) -> None:
        sut = Place('P1', [PlaceName('The Place')])
        self.assertCountEqual([], sut.links)

    def test_names(self) -> None:
        name = PlaceName('The Place')
        sut = Place('P1', [name])
        self.assertCountEqual([name], sut.names)

    def test_coordinates(self) -> None:
        name = PlaceName('The Place')
        sut = Place('P1', [name])
        coordinates = Point()
        sut.coordinates = coordinates
        self.assertEquals(coordinates, sut.coordinates)


class SubjectTest(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Subject.name(), str)
        self.assertNotEqual('', Subject.name)

    def test_label(self) -> None:
        sut = Subject()
        with Translations(NullTranslations()):
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class WitnessTest(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Witness.name(), str)
        self.assertNotEqual('', Witness.name)

    def test_label(self) -> None:
        sut = Witness()
        with Translations(NullTranslations()):
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class BeneficiaryTest(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Beneficiary.name(), str)
        self.assertNotEqual('', Beneficiary.name)

    def test_label(self) -> None:
        sut = Beneficiary()
        with Translations(NullTranslations()):
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class AttendeeTest(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Attendee.name(), str)
        self.assertNotEqual('', Attendee.name)

    def test_label(self) -> None:
        sut = Attendee()
        with Translations(NullTranslations()):
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class PresenceTest(TestCase):
    def test_person(self) -> None:
        person = Mock(Person)
        sut = Presence(person, Mock(PresenceRole), Mock(Event))
        self.assertEquals(person, sut.person)

    def test_event(self) -> None:
        role = Mock(PresenceRole)
        sut = Presence(Mock(Person), role, Mock(Event))
        self.assertEquals(role, sut.role)

    def test_role(self) -> None:
        event = Mock(Event)
        sut = Presence(Mock(Person), Mock(PresenceRole), event)
        self.assertEquals(event, sut.event)


class EventTypeTest(TestCase):
    def test_comes_before(self) -> None:
        self.assertIsInstance(EventType.comes_before(), set)

    def test_comes_after(self) -> None:
        self.assertIsInstance(EventType.comes_after(), set)


class UnknownEventTypeTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(UnknownEventType.name(), str)
        self.assertNotEqual('', UnknownEventType.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = UnknownEventType()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)

    def test_comes_before(self) -> None:
        self.assertIsInstance(UnknownEventType.comes_before(), set)

    def test_comes_after(self) -> None:
        self.assertIsInstance(UnknownEventType.comes_after(), set)


class PreBirthEventTypeTestCase(TestCase):
    def test_comes_before(self) -> None:
        self.assertIsInstance(PreBirthEventType.comes_before(), set)


class LifeEventTypeTestCase(TestCase):
    def test_comes_before(self) -> None:
        self.assertIsInstance(LifeEventType.comes_before(), set)

    def test_comes_after(self) -> None:
        self.assertIsInstance(LifeEventType.comes_after(), set)


class PostDeathEventTypeTestCase(TestCase):
    def test_comes_after(self) -> None:
        self.assertIsInstance(PostDeathEventType.comes_after(), set)


class BirthTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Birth.name(), str)
        self.assertNotEqual('', Birth.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Birth()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)

    def test_comes_before(self) -> None:
        self.assertIsInstance(Birth.comes_before(), set)


class BaptismTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Baptism.name(), str)
        self.assertNotEqual('', Baptism.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Baptism()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class AdoptionTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Adoption.name(), str)
        self.assertNotEqual('', Adoption.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Adoption()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class DeathTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Death.name(), str)
        self.assertNotEqual('', Death.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Death()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)

    def test_comes_after(self) -> None:
        self.assertIsInstance(Death.comes_after(), set)


class FuneralTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Funeral.name(), str)
        self.assertNotEqual('', Funeral.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Funeral()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)

    def test_comes_after(self) -> None:
        self.assertIsInstance(Funeral.comes_after(), set)


class FinalDispositionEventTypeTestCase(TestCase):
    def test_comes_after(self) -> None:
        self.assertIsInstance(FinalDispositionEventType.comes_after(), set)


class CremationTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Cremation.name(), str)
        self.assertNotEqual('', Cremation.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Cremation()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class BurialTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Burial.name(), str)
        self.assertNotEqual('', Burial.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Burial()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class WillTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Will.name(), str)
        self.assertNotEqual('', Will.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Will()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)

    def test_comes_after(self) -> None:
        self.assertIsInstance(Will.comes_after(), set)


class EngagementTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Engagement.name(), str)
        self.assertNotEqual('', Engagement.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Engagement()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)

    def test_comes_before(self) -> None:
        self.assertIsInstance(Engagement.comes_before(), set)


class MarriageTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Marriage.name(), str)
        self.assertNotEqual('', Marriage.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Marriage()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class MarriageAnnouncementTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(MarriageAnnouncement.name(), str)
        self.assertNotEqual('', MarriageAnnouncement.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = MarriageAnnouncement()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)

    def test_comes_before(self) -> None:
        self.assertIsInstance(MarriageAnnouncement.comes_before(), set)


class DivorceTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Divorce.name(), str)
        self.assertNotEqual('', Divorce.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Divorce()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)

    def test_comes_after(self) -> None:
        self.assertIsInstance(Divorce.comes_after(), set)


class DivorceAnnouncementTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(DivorceAnnouncement.name(), str)
        self.assertNotEqual('', DivorceAnnouncement.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = DivorceAnnouncement()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)

    def test_comes_before(self) -> None:
        self.assertIsInstance(DivorceAnnouncement.comes_before(), set)

    def test_comes_after(self) -> None:
        self.assertIsInstance(DivorceAnnouncement.comes_after(), set)


class ResidenceTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Residence.name(), str)
        self.assertNotEqual('', Residence.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Residence()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class ImmigrationTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Immigration.name(), str)
        self.assertNotEqual('', Immigration.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Immigration()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class EmigrationTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Emigration.name(), str)
        self.assertNotEqual('', Emigration.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Emigration()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class OccupationTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Occupation.name(), str)
        self.assertNotEqual('', Occupation.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Occupation()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class RetirementTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Retirement.name(), str)
        self.assertNotEqual('', Retirement.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Retirement()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class CorrespondenceTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Correspondence.name(), str)
        self.assertNotEqual('', Correspondence.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Correspondence()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class ConfirmationTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Confirmation.name(), str)
        self.assertNotEqual('', Confirmation.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Confirmation()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class MissingTestCase(TestCase):
    def test_name(self) -> None:
        self.assertIsInstance(Missing.name(), str)
        self.assertNotEqual('', Missing.name())

    def test_label(self) -> None:
        with Translations(NullTranslations()):
            sut = Missing()
            self.assertIsInstance(sut.label, str)
            self.assertNotEqual('', sut.label)


class EventTypeTypesTest(TestCase):
    def test(self) -> None:
        for event_type_type in EVENT_TYPE_TYPES:
            self.assertTrue(issubclass(event_type_type, EventType))


class EventTest(TestCase):
    def test_resource_type_name(self) -> None:
        self.assertIsInstance(Event.resource_type_name(), str)
        self.assertNotEqual('', Event.resource_type_name())

    def test_place(self) -> None:
        place = Place('1', [PlaceName('one')])
        sut = Event(Mock(EventType))
        sut.place = place
        self.assertEquals(place, sut.place)
        self.assertIn(sut, place.events)
        sut.place = None
        self.assertEquals(None, sut.place)
        self.assertNotIn(sut, place.events)

    def test_presences(self) -> None:
        person = Person('P1')
        sut = Event(Mock(EventType))
        presence = Presence(person, Subject(), sut)
        sut.presences.append(presence)
        self.assertCountEqual([presence], sut.presences)
        self.assertEquals(sut, presence.event)
        sut.presences.remove(presence)
        self.assertCountEqual([], sut.presences)
        self.assertIsNone(presence.event)

    def test_date(self) -> None:
        sut = Event(Mock(EventType))
        self.assertIsNone(sut.date)
        date = Mock(Date)
        sut.date = date
        self.assertEquals(date, sut.date)

    def test_files(self) -> None:
        sut = Event(Mock(EventType))
        self.assertCountEqual([], sut.files)

    def test_citations(self) -> None:
        sut = Event(Mock(EventType))
        self.assertCountEqual([], sut.citations)

    def test_description(self) -> None:
        sut = Event(Mock(EventType))
        self.assertIsNone(sut.description)

    def test_private(self) -> None:
        sut = Event(Mock(EventType))
        self.assertIsNone(sut.private)

    def test_type(self) -> None:
        event_type = Mock(EventType)
        sut = Event(event_type)
        self.assertEquals(event_type, sut.type)

    def test_associated_files(self) -> None:
        file1 = Mock(File)
        file2 = Mock(File)
        file3 = Mock(File)
        file4 = Mock(File)
        sut = Event(Mock(EventType))
        sut.files = [file1, file2, file1]
        citation = Mock(Citation)
        citation.associated_files = [file3, file4, file2]
        sut.citations = [citation]
        self.assertEquals([file1, file2, file3, file4], list(sut.associated_files))


class IdentifiableEventTest(TestCase):
    def test_id(self) -> None:
        event_id = 'E1'
        sut = IdentifiableEvent(event_id, Mock(EventType))
        self.assertEquals(event_id, sut.id)


class PersonNameTest(TestCase):
    def test_person(self) -> None:
        sut = PersonName('Janet', 'Not a Girl')
        person = Person('1')
        sut.person = person
        self.assertEquals(person, sut.person)
        self.assertCountEqual([sut], person.names)
        sut.person = None
        self.assertIsNone(sut.person)
        self.assertCountEqual([], person.names)

    def test_locale(self) -> None:
        sut = PersonName('Janet', 'Not a Girl')
        self.assertIsNone(sut.locale)

    def test_citations(self) -> None:
        sut = PersonName('Janet', 'Not a Girl')
        self.assertCountEqual([], sut.citations)

    def test_individual(self) -> None:
        individual = 'Janet'
        sut = PersonName(individual, 'Not a Girl')
        self.assertEquals(individual, sut.individual)

    def test_affiliation(self) -> None:
        affiliation = 'Not a Girl'
        sut = PersonName('Janet', affiliation)
        self.assertEquals(affiliation, sut.affiliation)

    @parameterized.expand([
        (True, PersonName('Janet', 'Not a Girl'), PersonName('Janet', 'Not a Girl')),
        (True, PersonName('Janet'), PersonName('Janet')),
        (True, PersonName(None, 'Not a Girl'), PersonName(None, 'Not a Girl')),
        (False, PersonName('Janet'), PersonName(None, 'Not a Girl')),
        (False, PersonName('Janet', 'Not a Girl'), None),
        (False, PersonName('Janet', 'Not a Girl'), True),
        (False, PersonName('Janet', 'Not a Girl'), 9),
        (False, PersonName('Janet', 'Not a Girl'), object()),
    ])
    def test_eq(self, expected: bool, left: PersonName, right: Any) -> None:
        self.assertEquals(expected, left == right)

    @parameterized.expand([
        (False, PersonName('Janet', 'Not a Girl'), PersonName('Janet', 'Not a Girl')),
        (True, PersonName('Janet', 'Not a Girl'), PersonName('Not a Girl', 'Janet')),
        (True, PersonName('Janet', 'Not a Girl'), None),
    ])
    def test_gt(self, expected: bool, left: PersonName, right: Any) -> None:
        self.assertEquals(expected, left > right)


class PersonTest(TestCase):
    def test_resource_type_name(self) -> None:
        self.assertIsInstance(Person.resource_type_name(), str)
        self.assertNotEqual('', Person.resource_type_name())

    def test_parents(self) -> None:
        sut = Person('1')
        parent = Person('2')
        sut.parents.append(parent)
        self.assertCountEqual([parent], sut.parents)
        self.assertCountEqual([sut], parent.children)
        sut.parents.remove(parent)
        self.assertCountEqual([], sut.parents)
        self.assertCountEqual([], parent.children)

    def test_children(self) -> None:
        sut = Person('1')
        child = Person('2')
        sut.children.append(child)
        self.assertCountEqual([child], sut.children)
        self.assertCountEqual([sut], child.parents)
        sut.children.remove(child)
        self.assertCountEqual([], sut.children)
        self.assertCountEqual([], child.parents)

    def test_presences(self) -> None:
        event = Event(Birth())
        sut = Person('1')
        presence = Presence(sut, Subject(), event)
        sut.presences.append(presence)
        self.assertCountEqual([presence], sut.presences)
        self.assertEquals(sut, presence.person)
        sut.presences.remove(presence)
        self.assertCountEqual([], sut.presences)
        self.assertIsNone(presence.person)

    def test_names(self) -> None:
        sut = Person('1')
        name = PersonName('Janet', 'Not a Girl')
        sut.names.append(name)
        self.assertCountEqual([name], sut.names)
        self.assertEquals(sut, name.person)
        sut.names.remove(name)
        self.assertCountEqual([], sut.names)
        self.assertIsNone(name.person)

    def test_id(self) -> None:
        person_id = 'P1'
        sut = Person(person_id)
        self.assertEquals(person_id, sut.id)

    def test_files(self) -> None:
        sut = Source()
        self.assertCountEqual([], sut.files)

    def test_citations(self) -> None:
        sut = Source()
        self.assertCountEqual([], sut.citations)

    def test_links(self) -> None:
        sut = Source()
        self.assertCountEqual([], sut.links)

    def test_private(self) -> None:
        sut = Event(Mock(EventType))
        self.assertIsNone(sut.private)

    def test_name_with_names(self) -> None:
        sut = Person('P1')
        name = PersonName()
        sut.names = [name, PersonName()]
        self.assertEquals(name, sut.name)

    def test_name_without_names(self) -> None:
        self.assertIsNone(Person('P1').name)

    def test_alternative_names(self) -> None:
        sut = Person('P1')
        name = PersonName('Janet', 'Not a Girl')
        alternative_name = PersonName('Janet', 'Still not a Girl')
        sut.names = [name, alternative_name]
        self.assertEquals([alternative_name], sut.alternative_names)

    def test_start(self) -> None:
        start = Event(Birth())
        sut = Person('P1')
        Presence(sut, Subject(), start)
        self.assertEquals(start, sut.start)

    def test_end(self) -> None:
        end = Event(Burial())
        sut = Person('P1')
        Presence(sut, Subject(), end)
        self.assertEquals(end, sut.end)

    def test_siblings_without_parents(self) -> None:
        sut = Person('person')
        self.assertCountEqual([], sut.siblings)

    def test_siblings_with_one_common_parent(self) -> None:
        sut = Person('1')
        sibling = Person('2')
        parent = Person('3')
        parent.children = [sut, sibling]
        self.assertCountEqual([sibling], sut.siblings)

    def test_siblings_with_multiple_common_parents(self) -> None:
        sut = Person('1')
        sibling = Person('2')
        parent = Person('3')
        parent.children = [sut, sibling]
        self.assertCountEqual([sibling], sut.siblings)

    def test_associated_files(self) -> None:
        file1 = Mock(File)
        file2 = Mock(File)
        file3 = Mock(File)
        file4 = Mock(File)
        file5 = Mock(File)
        file6 = Mock(File)
        sut = Person('1')
        sut.files = [file1, file2, file1]
        citation = Mock(Citation)
        citation.associated_files = [file3, file4, file2]
        name = PersonName()
        name.citations = [citation]
        sut.names = [name]
        event = Mock(Event)
        event.associated_files = [file5, file6, file4]
        Presence(sut, Subject(), event)
        self.assertEquals([file1, file2, file3, file4, file5, file6], list(sut.associated_files))


class ResourceTypesTest(TestCase):
    def test(self) -> None:
        for resource_type in RESOURCE_TYPES:
            self.assertTrue(issubclass(resource_type, Resource))



import json as stdjson
from tempfile import TemporaryDirectory, NamedTemporaryFile

from geopy import Point

from betty import json
from betty.ancestry import Place, Person, PlaceName, Link, Presence, Source, File, Note, PersonName, \
    IdentifiableEvent, IdentifiableSource, IdentifiableCitation, Subject, Birth, Enclosure
from betty.config import Configuration, LocaleConfiguration
from betty.json import JSONEncoder
from betty.locale import Date, DateRange
from betty.media_type import MediaType
from betty.site import Site
from betty.tests import TestCase


class JSONEncoderTest(TestCase):
    def assert_encodes(self, expected, data, schema_definition: str):
        with TemporaryDirectory() as output_directory:
            configuration = Configuration(
                output_directory, '')
            configuration.locales.clear()
            configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
            configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
            site = Site(configuration)
            encoded_data = stdjson.loads(stdjson.dumps(data, cls=JSONEncoder.get_factory(
                site, configuration.default_locale)))
            json.validate(encoded_data, schema_definition, site)
            self.assertEquals(expected, encoded_data)

    def test_coordinates_should_encode(self):
        latitude = 12.345
        longitude = -54.321
        coordinates = Point(latitude, longitude)
        expected = {
            '@context': {
                'latitude': 'https://schema.org/latitude',
                'longitude': 'https://schema.org/longitude',
            },
            '@type': 'https://schema.org/GeoCoordinates',
            'latitude': latitude,
            'longitude': longitude,
        }
        self.assert_encodes(expected, coordinates, 'coordinates')

    def test_place_should_encode_minimal(self):
        place_id = 'the_place'
        name = 'The Place'
        place = Place(place_id, [PlaceName(name)])
        expected = {
            '$schema': '/schema.json#/definitions/place',
            '@context': {
                'enclosedBy': 'https://schema.org/containedInPlace',
                'encloses': 'https://schema.org/containsPlace',
                'events': 'https://schema.org/event'
            },
            '@type': 'https://schema.org/Place',
            'id': place_id,
            'names': [
                {
                    'name': name,
                },
            ],
            'enclosedBy': [],
            'encloses': [],
            'events': [],
            'links': [
                {
                    'url': '/en/place/the_place/index.json',
                    'relationship': 'canonical',
                    'mediaType': 'application/json',
                },
                {
                    'url': '/nl/place/the_place/index.json',
                    'relationship': 'alternate',
                    'locale': 'nl-NL',
                },
                {
                    'url': '/en/place/the_place/index.html',
                    'relationship': 'alternate',
                    'mediaType': 'text/html',
                },
            ],
        }
        self.assert_encodes(expected, place, 'place')

    def test_place_should_encode_full(self):
        place_id = 'the_place'
        name = 'The Place'
        locale = 'nl-NL'
        latitude = 12.345
        longitude = -54.321
        coordinates = Point(latitude, longitude)
        place = Place(place_id, [PlaceName(name, locale)])
        place.coordinates = coordinates
        Enclosure(place, Place('the_enclosing_place', []))
        Enclosure(Place('the_enclosed_place', []), place)
        link = Link('https://example.com/the-place')
        link.label = 'The Place Online'
        place.links.add(link)
        place.events.append(IdentifiableEvent('E1', Birth()))
        expected = {
            '$schema': '/schema.json#/definitions/place',
            '@context': {
                'enclosedBy': 'https://schema.org/containedInPlace',
                'encloses': 'https://schema.org/containsPlace',
                'events': 'https://schema.org/event',
                'coordinates': 'https://schema.org/geo',
            },
            '@type': 'https://schema.org/Place',
            'id': place_id,
            'names': [
                {
                    'name': name,
                    'locale': 'nl-NL',
                },
            ],
            'events': [
                '/en/event/E1/index.json',
            ],
            'links': [
                {
                    'url': '/en/place/the_place/index.json',
                    'relationship': 'canonical',
                    'mediaType': 'application/json',
                },
                {
                    'url': '/nl/place/the_place/index.json',
                    'relationship': 'alternate',
                    'locale': 'nl-NL',
                },
                {
                    'url': '/en/place/the_place/index.html',
                    'relationship': 'alternate',
                    'mediaType': 'text/html',
                },
                {
                    'url': 'https://example.com/the-place',
                    'label': 'The Place Online',
                },
            ],
            'coordinates': {
                '@context': {
                    'latitude': 'https://schema.org/latitude',
                    'longitude': 'https://schema.org/longitude',
                },
                '@type': 'https://schema.org/GeoCoordinates',
                'latitude': latitude,
                'longitude': longitude,
            },
            'encloses': [
                '/en/place/the_enclosed_place/index.json',
            ],
            'enclosedBy': [
                '/en/place/the_enclosing_place/index.json',
            ],
        }
        self.assert_encodes(expected, place, 'place')

    def test_person_should_encode_minimal(self):
        person_id = 'the_person'
        person = Person(person_id)
        expected = {
            '$schema': '/schema.json#/definitions/person',
            '@context': {
                'parents': 'https://schema.org/parent',
                'children': 'https://schema.org/child',
                'siblings': 'https://schema.org/sibling',
            },
            '@type': 'https://schema.org/Person',
            'id': person_id,
            'names': [],
            'parents': [],
            'children': [],
            'siblings': [],
            'private': None,
            'presences': [],
            'citations': [],
            'links': [
                {
                    'url': '/en/person/the_person/index.json',
                    'relationship': 'canonical',
                    'mediaType': 'application/json',
                },
                {
                    'url': '/nl/person/the_person/index.json',
                    'relationship': 'alternate',
                    'locale': 'nl-NL',
                },
                {
                    'url': '/en/person/the_person/index.html',
                    'relationship': 'alternate',
                    'mediaType': 'text/html',
                },
            ],
        }
        self.assert_encodes(expected, person, 'person')

    def test_person_should_encode_full(self):
        parent_id = 'the_parent'
        parent = Person(parent_id)

        child_id = 'the_child'
        child = Person(child_id)

        sibling_id = 'the_sibling'
        sibling = Person(sibling_id)
        sibling.parents.append(parent)

        person_id = 'the_person'
        person_affiliation_name = 'Person'
        person_individual_name = 'The'
        person = Person(person_id)
        person.names.append(PersonName(person_individual_name, person_affiliation_name))
        person.parents.append(parent)
        person.children.append(child)
        person.private = False
        link = Link('https://example.com/the-person')
        link.label = 'The Person Online'
        person.links.add(link)
        person.citations.append(
            IdentifiableCitation('the_citation', Source('The Source')))
        Presence(person, Subject(), IdentifiableEvent('the_event', Birth()))

        expected = {
            '$schema': '/schema.json#/definitions/person',
            '@context': {
                'parents': 'https://schema.org/parent',
                'children': 'https://schema.org/child',
                'siblings': 'https://schema.org/sibling',
            },
            '@type': 'https://schema.org/Person',
            'id': person_id,
            'names': [
                {
                    '@context': {
                        'individual': 'https://schema.org/givenName',
                        'affiliation': 'https://schema.org/familyName',
                    },
                    'individual': person_individual_name,
                    'affiliation': person_affiliation_name,
                },
            ],
            'parents': [
                '/en/person/the_parent/index.json',
            ],
            'children': [
                '/en/person/the_child/index.json',
            ],
            'siblings': [
                '/en/person/the_sibling/index.json',
            ],
            'private': False,
            'presences': [
                {
                    '@context': {
                        'event': 'https://schema.org/performerIn',
                    },
                    'role': 'subject',
                    'event': '/en/event/the_event/index.json',
                },
            ],
            'citations': [
                '/en/citation/the_citation/index.json',
            ],
            'links': [
                {
                    'url': '/en/person/the_person/index.json',
                    'relationship': 'canonical',
                    'mediaType': 'application/json',
                },
                {
                    'url': '/nl/person/the_person/index.json',
                    'relationship': 'alternate',
                    'locale': 'nl-NL',
                },
                {
                    'url': '/en/person/the_person/index.html',
                    'relationship': 'alternate',
                    'mediaType': 'text/html',
                },
                {
                    'url': 'https://example.com/the-person',
                    'label': 'The Person Online',
                },
            ],
        }
        self.assert_encodes(expected, person, 'person')

    def test_file_should_encode_minimal(self):
        with NamedTemporaryFile() as f:
            file = File('the_file', f.name)
            expected = {
                '$schema': '/schema.json#/definitions/file',
                'id': 'the_file',
                'resources': [],
                'notes': [],
                'links': [
                    {
                        'url': '/en/file/the_file/index.json',
                        'relationship': 'canonical',
                        'mediaType': 'application/json',
                    },
                    {
                        'url': '/nl/file/the_file/index.json',
                        'relationship': 'alternate',
                        'locale': 'nl-NL',
                    },
                    {
                        'url': '/en/file/the_file/index.html',
                        'relationship': 'alternate',
                        'mediaType': 'text/html',
                    },
                ],
            }
            self.assert_encodes(expected, file, 'file')

    def test_file_should_encode_full(self):
        with NamedTemporaryFile() as f:
            file = File('the_file', f.name)
            file.media_type = MediaType('text/plain')
            file.notes.append(Note('The Note'))
            Person('the_person').files.append(file)
            expected = {
                '$schema': '/schema.json#/definitions/file',
                'id': 'the_file',
                'mediaType': 'text/plain',
                'resources': [
                    '/en/person/the_person/index.json',
                ],
                'notes': [
                    {
                        'text': 'The Note',
                    },
                ],
                'links': [
                    {
                        'url': '/en/file/the_file/index.json',
                        'relationship': 'canonical',
                        'mediaType': 'application/json',
                    },
                    {
                        'url': '/nl/file/the_file/index.json',
                        'relationship': 'alternate',
                        'locale': 'nl-NL',
                    },
                    {
                        'url': '/en/file/the_file/index.html',
                        'relationship': 'alternate',
                        'mediaType': 'text/html',
                    },
                ],
            }
            self.assert_encodes(expected, file, 'file')

    def test_event_should_encode_minimal(self):
        event = IdentifiableEvent('the_event', Birth())
        expected = {
            '$schema': '/schema.json#/definitions/event',
            '@type': 'https://schema.org/Event',
            'id': 'the_event',
            'type': 'birth',
            'presences': [],
            'citations': [],
            'links': [
                {
                    'url': '/en/event/the_event/index.json',
                    'relationship': 'canonical',
                    'mediaType': 'application/json',
                },
                {
                    'url': '/nl/event/the_event/index.json',
                    'relationship': 'alternate',
                    'locale': 'nl-NL',
                },
                {
                    'url': '/en/event/the_event/index.html',
                    'relationship': 'alternate',
                    'mediaType': 'text/html',
                },
            ],
        }
        self.assert_encodes(expected, event, 'event')

    def test_event_should_encode_full(self):
        event = IdentifiableEvent('the_event', Birth())
        event.date = DateRange(Date(2000, 1, 1), Date(2019, 12, 31))
        event.place = Place('the_place', [PlaceName('The Place')])
        Presence(Person('the_person'), Subject(), event)
        event.citations.append(
            IdentifiableCitation('the_citation', Source('The Source')))
        expected = {
            '$schema': '/schema.json#/definitions/event',
            '@context': {
                'place': 'https://schema.org/location',
            },
            '@type': 'https://schema.org/Event',
            'id': 'the_event',
            'type': 'birth',
            'presences': [
                {
                    '@context': {
                        'person': 'https://schema.org/actor',
                    },
                    'role': 'subject',
                    'person': '/en/person/the_person/index.json',
                },
            ],
            'citations': [
                '/en/citation/the_citation/index.json',
            ],
            'date': {
                'start': {
                    'year': 2000,
                    'month': 1,
                    'day': 1,
                },
                'end': {
                    'year': 2019,
                    'month': 12,
                    'day': 31,
                },
            },
            'place': '/en/place/the_place/index.json',
            'links': [
                {
                    'url': '/en/event/the_event/index.json',
                    'relationship': 'canonical',
                    'mediaType': 'application/json',
                },
                {
                    'url': '/nl/event/the_event/index.json',
                    'relationship': 'alternate',
                    'locale': 'nl-NL',
                },
                {
                    'url': '/en/event/the_event/index.html',
                    'relationship': 'alternate',
                    'mediaType': 'text/html',
                },
            ],
        }
        self.assert_encodes(expected, event, 'event')

    def test_source_should_encode_minimal(self):
        source = IdentifiableSource('the_source', 'The Source')
        expected = {
            '$schema': '/schema.json#/definitions/source',
            '@context': {
                'name': 'https://schema.org/name',
            },
            '@type': 'https://schema.org/Thing',
            'id': 'the_source',
            'name': 'The Source',
            'contains': [],
            'citations': [],
            'links': [
                {
                    'url': '/en/source/the_source/index.json',
                    'relationship': 'canonical',
                    'mediaType': 'application/json',
                },
                {
                    'url': '/nl/source/the_source/index.json',
                    'relationship': 'alternate',
                    'locale': 'nl-NL',
                },
                {
                    'url': '/en/source/the_source/index.html',
                    'relationship': 'alternate',
                    'mediaType': 'text/html',
                },
            ],
        }
        self.assert_encodes(expected, source, 'source')

    def test_source_should_encode_full(self):
        source = IdentifiableSource('the_source', 'The Source')
        source.author = 'The Author'
        source.publisher = 'The Publisher'
        source.date = Date(2000, 1, 1)
        source.contained_by = IdentifiableSource(
            'the_containing_source', 'The Containing Source')
        link = Link('https://example.com/the-source')
        link.label = 'The Source Online'
        source.links.add(link)
        source.contains.append(
            IdentifiableSource('the_contained_source', 'The Contained Source'))
        IdentifiableCitation('the_citation', source)
        expected = {
            '$schema': '/schema.json#/definitions/source',
            '@context': {
                'name': 'https://schema.org/name',
            },
            '@type': 'https://schema.org/Thing',
            'id': 'the_source',
            'name': 'The Source',
            'author': 'The Author',
            'publisher': 'The Publisher',
            'contains': [
                '/en/source/the_contained_source/index.json',
            ],
            'citations': [
                '/en/citation/the_citation/index.json',
            ],
            'containedBy': '/en/source/the_containing_source/index.json',
            'date': {
                'year': 2000,
                'month': 1,
                'day': 1,
            },
            'links': [
                {
                    'url': '/en/source/the_source/index.json',
                    'relationship': 'canonical',
                    'mediaType': 'application/json',
                },
                {
                    'url': '/nl/source/the_source/index.json',
                    'relationship': 'alternate',
                    'locale': 'nl-NL',
                },
                {
                    'url': '/en/source/the_source/index.html',
                    'relationship': 'alternate',
                    'mediaType': 'text/html',
                },
                {
                    'url': 'https://example.com/the-source',
                    'label': 'The Source Online',
                },
            ],
        }
        self.assert_encodes(expected, source, 'source')

    def test_citation_should_encode_minimal(self):
        citation = IdentifiableCitation('the_citation', Source('The Source'))
        expected = {
            '$schema': '/schema.json#/definitions/citation',
            '@type': 'https://schema.org/Thing',
            'id': 'the_citation',
            'facts': [],
            'links': [
                {
                    'url': '/en/citation/the_citation/index.json',
                    'relationship': 'canonical',
                    'mediaType': 'application/json',
                },
                {
                    'url': '/nl/citation/the_citation/index.json',
                    'relationship': 'alternate',
                    'locale': 'nl-NL',
                },
                {
                    'url': '/en/citation/the_citation/index.html',
                    'relationship': 'alternate',
                    'mediaType': 'text/html',
                },
            ],
        }
        self.assert_encodes(expected, citation, 'citation')

    def test_citation_should_encode_full(self):
        citation = IdentifiableCitation('the_citation', IdentifiableSource('the_source', 'The Source'))
        citation.description = 'The Source Description'
        citation.facts.append(IdentifiableEvent('the_event', Birth()))
        expected = {
            '$schema': '/schema.json#/definitions/citation',
            '@type': 'https://schema.org/Thing',
            'id': 'the_citation',
            'source': '/en/source/the_source/index.json',
            'facts': [
                '/en/event/the_event/index.json'
            ],
            'links': [
                {
                    'url': '/en/citation/the_citation/index.json',
                    'relationship': 'canonical',
                    'mediaType': 'application/json',
                },
                {
                    'url': '/nl/citation/the_citation/index.json',
                    'relationship': 'alternate',
                    'locale': 'nl-NL',
                },
                {
                    'url': '/en/citation/the_citation/index.html',
                    'relationship': 'alternate',
                    'mediaType': 'text/html',
                },
            ],
        }
        self.assert_encodes(expected, citation, 'citation')

    def test_link_should_encode_minimal(self) -> None:
        link = Link('https://example.com')
        expected = {
            'url': 'https://example.com',
        }
        self.assert_encodes(expected, link, 'link')

    def test_link_should_encode_full(self) -> None:
        link = Link('https://example.com')
        link.label = 'The Link'
        link.relationship = 'external'
        link.locale = 'nl-NL'
        link.media_type = MediaType('text/html')
        expected = {
            'url': 'https://example.com',
            'relationship': 'external',
            'label': 'The Link',
            'locale': 'nl-NL',
            'mediaType': 'text/html',
        }
        self.assert_encodes(expected, link, 'link')



from os.path import join, dirname, abspath
from tempfile import TemporaryDirectory
from typing import Optional

from parameterized import parameterized

from betty.ancestry import Ancestry, PersonName, Birth, Death, UnknownEventType
from betty.config import Configuration
from betty.asyncio import sync
from betty.locale import Date
from betty.parse import parse
from betty.path import rootname
from betty.plugin.gramps import parse_xml, Gramps, parse_gpkg, parse_gramps
from betty.site import Site
from betty.tests import TestCase


class ParseGrampsTest(TestCase):
    pass


class ParseGpkgTest(TestCase):
    pass


class ParseXmlTest(TestCase):
    @classmethod
    @sync
    async def setUpClass(cls) -> None:
        TestCase.setUpClass()
        # @todo Convert each test method to use self._parse(), so we can remove this shared XML file.
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(output_directory_path, 'https://example.com')
            async with Site(configuration) as site:
                cls.ancestry = site.ancestry
                xml_file_path = join(dirname(abspath(__file__)), 'assets', 'data.xml')
                with open(xml_file_path) as f:
                    parse_xml(site, f.read(), rootname(xml_file_path))

    @sync
    async def _parse(self, xml: str) -> Ancestry:
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(output_directory_path, 'https://example.com')
            async with Site(configuration) as site:
                with TemporaryDirectory() as tree_directory_path:
                    parse_xml(site, xml.strip(), tree_directory_path)
                    return site.ancestry

    def _parse_partial(self, xml: str) -> Ancestry:
        return self._parse("""
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE database PUBLIC "-//Gramps//DTD Gramps XML 1.7.1//EN"
"http://gramps-project.org/xml/1.7.1/grampsxml.dtd">
<database xmlns="http://gramps-project.org/xml/1.7.1/">
  <header>
    <created date="2019-03-09" version="4.2.8"/>
    <researcher>
    </researcher>
  </header>
  %s
</database>
""" % xml)

    @sync
    async def test_parse_xml_with_string(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(output_directory_path, 'https://example.com')
            async with Site(configuration) as site:
                gramps_file_path = join(dirname(abspath(__file__)), 'assets', 'minimal.xml')
                with open(gramps_file_path) as f:
                    parse_xml(site, f.read(), rootname(gramps_file_path))

    @sync
    async def test_parse_xml_with_file_path(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(output_directory_path, 'https://example.com')
            async with Site(configuration) as site:
                gramps_file_path = join(dirname(abspath(__file__)), 'assets', 'minimal.xml')
                parse_xml(site, gramps_file_path, rootname(gramps_file_path))

    @sync
    async def test_parse_gramps(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(output_directory_path, 'https://example.com')
            async with Site(configuration) as site:
                gramps_file_path = join(dirname(abspath(__file__)), 'assets', 'minimal.gramps')
                parse_gramps(site, gramps_file_path)

    @sync
    async def test_parse_gpkg(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(output_directory_path, 'https://example.com')
            async with Site(configuration) as site:
                gramps_file_path = join(dirname(abspath(__file__)), 'assets', 'minimal.gpkg')
                parse_gpkg(site, gramps_file_path)

    def test_place_should_include_name(self):
        place = self.ancestry.places['P0000']
        names = place.names
        self.assertEquals(1, len(names))
        name = names[0]
        self.assertEquals('Amsterdam', name.name)

    def test_place_should_include_coordinates(self):
        place = self.ancestry.places['P0000']
        self.assertAlmostEquals(52.366667, place.coordinates.latitude)
        self.assertAlmostEquals(4.9, place.coordinates.longitude)

    def test_place_should_include_events(self):
        place = self.ancestry.places['P0000']
        event = self.ancestry.events['E0000']
        self.assertIn(event, place.events)

    def test_person_should_include_name(self):
        person = self.ancestry.people['I0000']
        expected = PersonName('Jane', 'Doe')
        self.assertEquals(expected, person.name)

    def test_person_should_include_alternative_names(self):
        person = self.ancestry.people['I0000']
        expected = [
            PersonName('Jane', 'Doh'),
            PersonName('Jen', 'Van Doughie'),
        ]
        self.assertEquals(expected, person.alternative_names)

    def test_person_should_include_birth(self):
        person = self.ancestry.people['I0000']
        self.assertEquals('E0000', person.start.id)

    def test_person_should_include_death(self):
        person = self.ancestry.people['I0003']
        self.assertEquals('E0002', person.end.id)

    def test_person_should_be_private(self):
        person = self.ancestry.people['I0003']
        self.assertTrue(person.private)

    def test_person_should_not_be_private(self):
        person = self.ancestry.people['I0002']
        self.assertFalse(person.private)

    def test_person_should_include_citation(self):
        person = self.ancestry.people['I0000']
        citation = self.ancestry.citations['C0000']
        self.assertIn(citation, person.citations)

    def test_family_should_set_parents(self):
        expected_parents = [self.ancestry.people['I0002'],
                            self.ancestry.people['I0003']]
        children = [self.ancestry.people['I0000'],
                    self.ancestry.people['I0001']]
        for child in children:
            self.assertCountEqual(expected_parents, child.parents)

    def test_family_should_set_children(self):
        parents = [self.ancestry.people['I0002'],
                   self.ancestry.people['I0003']]
        expected_children = [self.ancestry.people['I0000'],
                             self.ancestry.people['I0001']]
        for parent in parents:
            self.assertCountEqual(expected_children, parent.children)

    def test_event_should_be_birth(self):
        self.assertIsInstance(self.ancestry.events['E0000'].type, Birth)

    def test_event_should_be_death(self):
        self.assertIsInstance(self.ancestry.events['E0002'].type, Death)

    def test_event_should_parse_unknown(self):
        ancestry = self._parse_partial("""
<events>
    <event handle="_e7692ea23775e80643fe4fcf91" change="1590243374" id="E0000">
        <type>SomeEventThatIUsedToKnow</type>
        <dateval val="0000-00-00" quality="calculated"/>
    </event>
</events>
""")
        self.assertIsInstance(ancestry.events['E0000'].type, UnknownEventType)

    def test_event_should_include_place(self):
        event = self.ancestry.events['E0000']
        place = self.ancestry.places['P0000']
        self.assertEquals(place, event.place)

    def test_event_should_include_date(self):
        event = self.ancestry.events['E0000']
        self.assertEquals(1970, event.date.year)
        self.assertEquals(1, event.date.month)
        self.assertEquals(1, event.date.day)

    def test_event_should_include_people(self):
        event = self.ancestry.events['E0000']
        expected_people = [self.ancestry.people['I0000']]
        self.assertCountEqual(
            expected_people, [presence.person for presence in event.presences])

    def test_event_should_include_description(self):
        event = self.ancestry.events['E0008']
        self.assertEquals('Something happened!', event.description)

    @parameterized.expand([
        (Date(), '0000-00-00'),
        (Date(None, None, 1), '0000-00-01'),
        (Date(None, 1), '0000-01-00'),
        (Date(None, 1, 1), '0000-01-01'),
        (Date(1970), '1970-00-00'),
        (Date(1970, None, 1), '1970-00-01'),
        (Date(1970, 1), '1970-01-00'),
        (Date(1970, 1, 1), '1970-01-01'),
    ])
    def test_date_should_parse_parts(self, expected: Date, dateval_val: str):
        ancestry = self._parse_partial("""
<events>
    <event handle="_e7692ea23775e80643fe4fcf91" change="1590243374" id="E0000">
        <type>Birth</type>
        <dateval val="%s" quality="calculated"/>
    </event>
</events>
""" % dateval_val)
        self.assertEquals(expected, ancestry.events['E0000'].date)

    def test_date_should_ignore_calendar_format(self):
        self.assertIsNone(self.ancestry.events['E0005'].date)

    def test_date_should_parse_before(self):
        date = self.ancestry.events['E0003'].date
        self.assertIsNone(date.start)
        self.assertEquals(1970, date.end.year)
        self.assertEquals(1, date.end.month)
        self.assertEquals(1, date.end.day)
        self.assertTrue(date.end_is_boundary)
        self.assertFalse(date.end.fuzzy)

    def test_date_should_parse_after(self):
        date = self.ancestry.events['E0004'].date
        self.assertIsNone(date.end)
        self.assertEquals(1970, date.start.year)
        self.assertEquals(1, date.start.month)
        self.assertEquals(1, date.start.day)
        self.assertTrue(date.start_is_boundary)
        self.assertFalse(date.start.fuzzy)

    def test_date_should_parse_calculated(self):
        ancestry = self._parse_partial("""
<events>
    <event handle="_e7692ea23775e80643fe4fcf91" change="1590243374" id="E0000">
        <type>Birth</type>
        <dateval val="1970-01-01" quality="calculated"/>
    </event>
</events>
""")
        date = ancestry.events['E0000'].date
        self.assertEquals(1970, date.year)
        self.assertEquals(1, date.month)
        self.assertEquals(1, date.day)
        self.assertFalse(date.fuzzy)

    def test_date_should_parse_estimated(self):
        ancestry = self._parse_partial("""
<events>
    <event handle="_e7692ea23775e80643fe4fcf91" change="1590243374" id="E0000">
        <type>Birth</type>
        <dateval val="1970-01-01" quality="estimated"/>
    </event>
</events>
""")
        date = ancestry.events['E0000'].date
        self.assertEquals(1970, date.year)
        self.assertEquals(1, date.month)
        self.assertEquals(1, date.day)
        self.assertTrue(date.fuzzy)

    def test_date_should_parse_about(self):
        date = self.ancestry.events['E0007'].date
        self.assertEquals(1970, date.year)
        self.assertEquals(1, date.month)
        self.assertEquals(1, date.day)
        self.assertTrue(date.fuzzy)

    def test_daterange_should_parse(self):
        ancestry = self._parse_partial("""
<events>
    <event handle="_e7692ea23775e80643fe4fcf91" change="1590243374" id="E0000">
        <type>Birth</type>
        <daterange start="1970-01-01" stop="1999-12-31"/>
    </event>
</events>
""")
        date = ancestry.events['E0000'].date
        self.assertEquals(1970, date.start.year)
        self.assertEquals(1, date.start.month)
        self.assertEquals(1, date.start.day)
        self.assertFalse(date.start.fuzzy)
        self.assertTrue(date.start_is_boundary)
        self.assertEquals(1999, date.end.year)
        self.assertEquals(12, date.end.month)
        self.assertEquals(31, date.end.day)
        self.assertTrue(date.end_is_boundary)
        self.assertFalse(date.end.fuzzy)

    def test_daterange_should_parse_calculated(self):
        ancestry = self._parse_partial("""
<events>
    <event handle="_e7692ea23775e80643fe4fcf91" change="1590243374" id="E0000">
        <type>Birth</type>
        <daterange start="1970-01-01" stop="1999-12-31" quality="calculated"/>
    </event>
</events>
""")
        date = ancestry.events['E0000'].date
        self.assertFalse(date.start.fuzzy)
        self.assertFalse(date.end.fuzzy)

    def test_daterange_should_parse_estimated(self):
        ancestry = self._parse_partial("""
<events>
    <event handle="_e7692ea23775e80643fe4fcf91" change="1590243374" id="E0000">
        <type>Birth</type>
        <daterange start="1970-01-01" stop="1999-12-31" quality="estimated"/>
    </event>
</events>
""")
        date = ancestry.events['E0000'].date
        self.assertTrue(date.start.fuzzy)
        self.assertTrue(date.end.fuzzy)

    def test_datespan_should_parse(self):
        ancestry = self._parse_partial("""
<events>
    <event handle="_e7692ea23775e80643fe4fcf91" change="1590243374" id="E0000">
        <type>Birth</type>
        <datespan start="1970-01-01" stop="1999-12-31"/>
    </event>
</events>
""")
        date = ancestry.events['E0000'].date
        self.assertEquals(1970, date.start.year)
        self.assertEquals(1, date.start.month)
        self.assertEquals(1, date.start.day)
        self.assertFalse(date.start.fuzzy)
        self.assertEquals(1999, date.end.year)
        self.assertEquals(12, date.end.month)
        self.assertEquals(31, date.end.day)
        self.assertFalse(date.end.fuzzy)

    def test_datespan_should_parse_calculated(self):
        ancestry = self._parse_partial("""
<events>
    <event handle="_e7692ea23775e80643fe4fcf91" change="1590243374" id="E0000">
        <type>Birth</type>
        <datespan start="1970-01-01" stop="1999-12-31" quality="calculated"/>
    </event>
</events>
""")
        date = ancestry.events['E0000'].date
        self.assertFalse(date.start.fuzzy)
        self.assertFalse(date.end.fuzzy)

    def test_datespan_should_parse_estimated(self):
        ancestry = self._parse_partial("""
<events>
    <event handle="_e7692ea23775e80643fe4fcf91" change="1590243374" id="E0000">
        <type>Birth</type>
        <datespan start="1970-01-01" stop="1999-12-31" quality="estimated"/>
    </event>
</events>
""")
        date = ancestry.events['E0000'].date
        self.assertTrue(date.start.fuzzy)
        self.assertTrue(date.end.fuzzy)

    def test_source_from_repository_should_include_name(self):
        source = self.ancestry.sources['R0000']
        self.assertEquals('Library of Alexandria', source.name)

    def test_source_from_repository_should_include_link(self):
        links = self.ancestry.sources['R0000'].links
        self.assertEquals(1, len(links))
        link = list(links)[0]
        self.assertEquals('https://alexandria.example.com', link.url)
        self.assertEquals('Library of Alexandria Catalogue', link.label)

    def test_source_from_source_should_include_title(self):
        source = self.ancestry.sources['S0000']
        self.assertEquals('A Whisper', source.name)

    def test_source_from_source_should_include_author(self):
        source = self.ancestry.sources['S0000']
        self.assertEquals('A Little Birdie', source.author)

    def test_source_from_source_should_include_publisher(self):
        source = self.ancestry.sources['S0000']
        self.assertEquals('Somewhere over the rainbow', source.publisher)

    def test_source_from_source_should_include_repository(self):
        source = self.ancestry.sources['S0000']
        containing_source = self.ancestry.sources['R0000']
        self.assertEquals(containing_source, source.contained_by)

    @parameterized.expand([
        (True, 'private'),
        (False, 'public'),
        (None, 'publi'),
        (None, 'privat'),
    ])
    def test_person_should_include_privacy_from_attribute(self, expected: Optional[bool], attribute_value: str) -> None:
        ancestry = self._parse_partial("""
<people>
    <person handle="_e1dd3ac2fa22e6fefa18f738bdd" change="1552126811" id="I0000">
        <gender>U</gender>
        <attribute type="betty:privacy" value="%s"/>
    </person>
</people>
""" % attribute_value)
        person = ancestry.people['I0000']
        self.assertEquals(expected, person.private)

    @parameterized.expand([
        (True, 'private'),
        (False, 'public'),
        (None, 'publi'),
        (None, 'privat'),
    ])
    def test_event_should_include_privacy_from_attribute(self, expected: Optional[bool], attribute_value: str) -> None:
        ancestry = self._parse_partial("""
<events>
    <event handle="_e1dd3ac2fa22e6fefa18f738bdd" change="1552126811" id="E0000">
        <type>Birth</type>
        <attribute type="betty:privacy" value="%s"/>
    </event>
</events>
""" % attribute_value)
        event = ancestry.events['E0000']
        self.assertEquals(expected, event.private)

    @parameterized.expand([
        (True, 'private'),
        (False, 'public'),
        (None, 'publi'),
        (None, 'privat'),
    ])
    def test_file_should_include_privacy_from_attribute(self, expected: Optional[bool], attribute_value: str) -> None:
        ancestry = self._parse_partial("""
<objects>
    <object handle="_e66f421249f3e9ebf6744d3b11d" change="1583534526" id="O0000">
        <file src="/tmp/file.txt" mime="text/plain" checksum="d41d8cd98f00b204e9800998ecf8427e" description="file"/>
        <attribute type="betty:privacy" value="%s"/>
    </object>
</objects>
""" % attribute_value)
        file = ancestry.files['O0000']
        self.assertEquals(expected, file.private)

    @parameterized.expand([
        (True, 'private'),
        (False, 'public'),
        (None, 'publi'),
        (None, 'privat'),
    ])
    def test_source_from_source_should_include_privacy_from_attribute(self, expected: Optional[bool], attribute_value: str) -> None:
        ancestry = self._parse_partial("""
<sources>
    <source handle="_e1dd686b04813540eb3503a342b" change="1558277217" id="S0000">
        <stitle>A Whisper</stitle>
        <srcattribute type="betty:privacy" value="%s"/>
    </source>
</sources>
""" % attribute_value)
        source = ancestry.sources['S0000']
        self.assertEquals(expected, source.private)

    @parameterized.expand([
        (True, 'private'),
        (False, 'public'),
        (None, 'publi'),
        (None, 'privat'),
    ])
    def test_citation_should_include_privacy_from_attribute(self, expected: Optional[bool], attribute_value: str) -> None:
        ancestry = self._parse_partial("""
<citations>
    <citation handle="_e2c25a12a097a0b24bd9eae5090" change="1558277266" id="C0000">
        <confidence>2</confidence>
        <sourceref hlink="_e1dd686b04813540eb3503a342b"/>
        <srcattribute type="betty:privacy" value="%s"/>
    </citation>
</citations>
<sources>
    <source handle="_e1dd686b04813540eb3503a342b" change="1558277217" id="S0000">
        <stitle>A Whisper</stitle>
    </source>
</sources>
""" % attribute_value)
        citation = ancestry.citations['C0000']
        self.assertEquals(expected, citation.private)


class GrampsTest(TestCase):
    @sync
    async def test_parse_event(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.plugins[Gramps] = {
                'file': join(dirname(abspath(__file__)), 'assets', 'minimal.gpkg')
            }
            async with Site(configuration) as site:
                await parse(site)
            self.assertEquals(
                'Dough', site.ancestry.people['I0000'].name.affiliation)
            self.assertEquals(
                'Janet', site.ancestry.people['I0000'].name.individual)
            self.assertEquals(
                '1px', site.ancestry.files['O0000'].description)



from betty.asyncio import sync
from betty.tests import TestCase


class SyncTest(TestCase):
    def test_call_coroutinefunction_should_return_result(self) -> None:
        expected = 'Hello, oh asynchronous, world!'

        async def _async():
            return expected
        actual = sync(_async())
        self.assertEqual(expected, actual)

    def test_call_decorated_coroutinefunction_should_return_result(self) -> None:
        expected = 'Hello, oh asynchronous, world!'

        @sync
        async def _async():
            return expected
        actual = _async()
        self.assertEqual(expected, actual)

    def test_call_decorated_function_should_return_result(self) -> None:
        expected = 'Hello, oh asynchronous, world!'

        @sync
        def _sync():
            return expected
        actual = _sync()
        self.assertEqual(expected, actual)

    def test_call_decorated_callable_method_should_return_result(self) -> None:
        expected = 'Hello, oh asynchronous, world!'

        class _Sync:
            @sync
            def __call__(self, *args, **kwargs):
                return expected
        actual = _Sync()()
        self.assertEqual(expected, actual)

    def test_call_decorated_callable_coroutinemethod_should_return_result(self) -> None:
        expected = 'Hello, oh asynchronous, world!'

        class _Sync:
            @sync
            async def __call__(self, *args, **kwargs):
                return expected
        actual = _Sync()()
        self.assertEqual(expected, actual)

    def test_call_wrapped_callable_object_should_return_result(self) -> None:
        expected = 'Hello, oh asynchronous, world!'

        class _Sync:
            def __call__(self, *args, **kwargs):
                return expected
        actual = sync(_Sync())()
        self.assertEqual(expected, actual)

    def test_call_wrapped_coroutinecallable_object_should_return_result(self) -> None:
        expected = 'Hello, oh asynchronous, world!'

        class _Sync:
            async def __call__(self, *args, **kwargs):
                return expected
        actual = sync(_Sync())()
        self.assertEqual(expected, actual)

    def test_unsychronizable(self) -> None:
        with self.assertRaises(ValueError):
            sync(True)



import json
import sys
import unittest
from os import path
from tempfile import TemporaryDirectory

import html5lib
import jsonschema
import requests
from requests import Response

from betty import generate
from betty.config import from_file
from betty.asyncio import sync
from betty.plugin.nginx.serve import DockerizedNginxServer
from betty.serve import Server
from betty.site import Site
from betty.tests import TestCase


@unittest.skipIf(sys.platform in {'darwin', 'win32'}, 'Mac OS and Windows do not natively support Docker.')
class NginxTest(TestCase):
    class NginxTestServer(Server):
        def __init__(self, configuration_template_file_path: str):
            with open(path.join(path.dirname(__file__), 'test_integration_assets', configuration_template_file_path)) as f:
                self._configuration = from_file(f)
            self._output_directory = None
            self._server = None

        async def start(self) -> None:
            self._output_directory = TemporaryDirectory()
            self._configuration.output_directory_path = self._output_directory.name
            site = Site(self._configuration)
            async with site:
                await generate.generate(site)
            self._server = DockerizedNginxServer(site)
            await self._server.start()

        async def stop(self) -> None:
            await self._server.stop()
            self._output_directory.cleanup()

        @property
        def public_url(self) -> str:
            return self._server.public_url

    def assert_betty_html(self, response: Response) -> None:
        self.assertEquals('text/html', response.headers['Content-Type'])
        parser = html5lib.HTMLParser()
        parser.parse(response.text)
        self.assertIn('Betty', response.text)

    def assert_betty_json(self, response: Response) -> None:
        self.assertEquals('application/json', response.headers['Content-Type'])
        data = response.json()
        with open(path.join(path.dirname(path.dirname(path.dirname(path.dirname(__file__)))), 'assets', 'public',
                            'static', 'schema.json')) as f:
            jsonschema.validate(data, json.load(f))

    @sync
    async def test_front_page(self):
        async with self.NginxTestServer('betty-monolingual.json') as server:
            response = requests.get(server.public_url)
            self.assertEquals(200, response.status_code)
            self.assert_betty_html(response)

    @sync
    async def test_default_html_404(self):
        async with self.NginxTestServer('betty-monolingual.json') as server:
            response = requests.get('%s/non-existent' % server.public_url)
            self.assertEquals(404, response.status_code)
            self.assert_betty_html(response)

    @sync
    async def test_negotiated_json_404(self):
        async with self.NginxTestServer('betty-monolingual-content-negotiation.json') as server:
            response = requests.get('%s/non-existent' % server.public_url, headers={
                'Accept': 'application/json',
            })
            self.assertEquals(404, response.status_code)
            self.assert_betty_json(response)

    @sync
    async def test_default_localized_front_page(self):
        async with self.NginxTestServer('betty-multilingual.json') as server:
            response = requests.get(server.public_url)
            self.assertEquals(200, response.status_code)
            self.assertEquals('en', response.headers['Content-Language'])
            self.assertEquals('%s/en/' % server.public_url, response.url)
            self.assert_betty_html(response)

    @sync
    async def test_explicitly_localized_404(self):
        async with self.NginxTestServer('betty-multilingual.json') as server:
            response = requests.get('%s/nl/non-existent' % server.public_url)
            self.assertEquals(404, response.status_code)
            self.assertEquals('nl', response.headers['Content-Language'])
            self.assert_betty_html(response)

    @sync
    async def test_negotiated_localized_front_page(self):
        async with self.NginxTestServer('betty-multilingual-content-negotiation.json') as server:
            response = requests.get(server.public_url, headers={
                'Accept-Language': 'nl-NL',
            })
            self.assertEquals(200, response.status_code)
            self.assertEquals('nl', response.headers['Content-Language'])
            self.assertEquals('%s/nl/' % server.public_url, response.url)
            self.assert_betty_html(response)

    @sync
    async def test_negotiated_localized_default_html_404(self):
        async with self.NginxTestServer('betty-multilingual-content-negotiation.json') as server:
            response = requests.get('%s/non-existent' % server.public_url, headers={
                'Accept-Language': 'nl-NL',
            })
            self.assertEquals(404, response.status_code)
            self.assertEquals('nl', response.headers['Content-Language'])
            self.assert_betty_html(response)

    @sync
    async def test_negotiated_localized_negotiated_json_404(self):
        async with self.NginxTestServer('betty-multilingual-content-negotiation.json') as server:
            response = requests.get('%s/non-existent' % server.public_url, headers={
                'Accept': 'application/json',
                'Accept-Language': 'nl-NL',
            })
            self.assertEquals(404, response.status_code)
            self.assert_betty_json(response)

    @sync
    async def test_default_html_resource(self):
        async with self.NginxTestServer('betty-monolingual-content-negotiation.json') as server:
            response = requests.get('%s/place/' % server.public_url)
            self.assertEquals(200, response.status_code)
            self.assert_betty_html(response)

    @sync
    async def test_negotiated_html_resource(self):
        async with self.NginxTestServer('betty-monolingual-content-negotiation.json') as server:
            response = requests.get('%s/place/' % server.public_url, headers={
                'Accept': 'text/html',
            })
            self.assertEquals(200, response.status_code)
            self.assert_betty_html(response)

    @sync
    async def test_negotiated_json_resource(self):
        async with self.NginxTestServer('betty-monolingual-content-negotiation.json') as server:
            response = requests.get('%s/place/' % server.public_url, headers={
                'Accept': 'application/json',
            })
            self.assertEquals(200, response.status_code)
            self.assert_betty_json(response)

    @sync
    async def test_default_html_static_resource(self):
        async with self.NginxTestServer('betty-multilingual-content-negotiation.json') as server:
            response = requests.get('%s/api/' % server.public_url)
            self.assertEquals(200, response.status_code)
            self.assert_betty_html(response)

    @sync
    async def test_negotiated_html_static_resource(self):
        async with self.NginxTestServer('betty-multilingual-content-negotiation.json') as server:
            response = requests.get('%s/api/' % server.public_url, headers={
                'Accept': 'text/html',
            })
            self.assertEquals(200, response.status_code)
            self.assert_betty_html(response)

    @sync
    async def test_negotiated_json_static_resource(self):
        async with self.NginxTestServer('betty-multilingual-content-negotiation.json') as server:
            response = requests.get('%s/api/' % server.public_url, headers={
                'Accept': 'application/json',
            })
            self.assertEquals(200, response.status_code)
            self.assert_betty_json(response)
            # Assert this is the exact JSON resource we are looking for.
            with open(path.join(path.dirname(__file__), 'test_integration_assets', 'schema.json')) as f:
                schema = json.load(f)
            jsonschema.validate(response.json(), schema)



from os import path
from tempfile import TemporaryDirectory

from parameterized import parameterized
from voluptuous import Invalid

from betty.tests import TestCase
from betty.voluptuous import Path, Importable


class PathTest(TestCase):
    @parameterized.expand([
        (3,),
        ({},),
        (True,),
    ])
    def test_with_invalid_type_should_raise_invalid(self, path_value):
        with self.assertRaises(Invalid):
            Path()(path_value)

    def test_with_absolute_path(self):
        with TemporaryDirectory() as directory_path:
            self.assertEqual(directory_path, Path()(directory_path))

    def test_with_expanduser(self):
        self.assertEqual(path.join(path.expanduser('~'), 'foo', 'bar'), Path()(path.join(path.expanduser('~'), 'foo', 'bar')))

    def test_with_relative_path_made_absolute(self):
        with TemporaryDirectory() as directory_path:
            self.assertEqual(path.join(directory_path, 'sibling'), Path()(path.join(directory_path, 'child', '..', 'sibling', '.')))


class ImportableTest(TestCase):
    @parameterized.expand([
        (3,),
        ({},),
        (True,),
    ])
    def test_with_invalid_type_should_raise_invalid(self, importable_value):
        with self.assertRaises(Invalid):
            Importable()(importable_value)

    def test_with_unknown_module_should_raise_invalid(self):
        with self.assertRaises(Invalid):
            Importable()('foo.bar.Baz')

    def test_with_unknown_type_should_raise_invalid(self):
        with self.assertRaises(Invalid):
            Importable()('%s.Foo' % self.__module__)

    def test_with_importable_should_return(self):
        self.assertEqual(self.__class__, Importable()('%s.%s' % (self.__module__, self.__class__.__name__)))



import json as stdjson
from os import path
from typing import Dict, Any, Union

import jsonschema
from geopy import Point
from jsonschema import RefResolver

from betty.ancestry import Place, Person, PlaceName, Event, Described, HasLinks, HasCitations, Link, Dated, File, \
    Note, PersonName, IdentifiableEvent, Identifiable, IdentifiableSource, IdentifiableCitation, HasMediaType, Resource, \
    PresenceRole, EventType
from betty.locale import Date, DateRange, Localized
from betty.media_type import MediaType
from betty.plugin.deriver import DerivedEvent
from betty.site import Site


def validate(data: Any, schema_definition: str, site: Site) -> None:
    with open(path.join(path.dirname(__file__), 'assets', 'public', 'static', 'schema.json')) as f:
        schema = stdjson.load(f)
    # @todo Can we set the schema ID somehow without making the entire JSON schema file a Jinja2 template?
    schema_id = site.static_url_generator.generate('schema.json', absolute=True)
    schema['$id'] = schema_id
    ref_resolver = RefResolver(schema_id, schema)
    jsonschema.validate(
        data, schema['definitions'][schema_definition], resolver=ref_resolver)


class JSONEncoder(stdjson.JSONEncoder):
    def __init__(self, site: Site, locale: str, *args, **kwargs):
        stdjson.JSONEncoder.__init__(self, *args, **kwargs)
        self._site = site
        self._locale = locale
        self._mappers = {
            PlaceName: self._encode_localized_name,
            Place: self._encode_place,
            Point: self._encode_coordinates,
            Person: self._encode_person,
            PersonName: self._encode_person_name,
            File: self._encode_file,
            DerivedEvent: self._encode_event,
            IdentifiableEvent: self._encode_identifiable_event,
            EventType: self._encode_event_type,
            PresenceRole: self._encode_presence_role,
            Date: self._encode_date,
            DateRange: self._encode_date_range,
            IdentifiableCitation: self._encode_identifiable_citation,
            IdentifiableSource: self._encode_identifiable_source,
            Link: self._encode_link,
            Note: self._encode_note,
            MediaType: self._encode_media_type,
        }

    @classmethod
    def get_factory(cls, site: Site, locale: str):
        return lambda *args, **kwargs: cls(site, locale, *args, **kwargs)

    def default(self, o):
        for mapper_type in self._mappers:
            if isinstance(o, mapper_type):
                return self._mappers[mapper_type](o)
        stdjson.JSONEncoder.default(self, o)

    def _generate_url(self, resource: Any, media_type='application/json', locale=None):
        locale = self._locale if locale is None else locale
        return self._site.localized_url_generator.generate(resource, media_type, locale=locale)

    def _encode_schema(self, encoded: Dict, defintion: str) -> None:
        encoded['$schema'] = self._site.static_url_generator.generate(
            'schema.json#/definitions/%s' % defintion)

    def _encode_identifiable_resource(self, encoded: Dict, resource: Union[Identifiable, Resource]) -> None:
        if 'links' not in encoded:
            encoded['links'] = []

        canonical = Link(self._generate_url(resource))
        canonical.relationship = 'canonical'
        canonical.media_type = 'application/json'
        encoded['links'].append(canonical)

        for locale in self._site.configuration.locales:
            if locale == self._locale:
                continue
            translation = Link(self._generate_url(resource, locale=locale))
            translation.relationship = 'alternate'
            translation.locale = locale
            encoded['links'].append(translation)

        html = Link(self._generate_url(resource, media_type='text/html'))
        html.relationship = 'alternate'
        html.media_type = 'text/html'
        encoded['links'].append(html)

    def _encode_described(self, encoded: Dict, described: Described) -> None:
        if described.description is not None:
            encoded['description'] = described.description
            encoded.update({
                '@context': {},
            })
            encoded['@context']['description'] = 'https://schema.org/description'

    def _encode_dated(self, encoded: Dict, dated: Dated) -> None:
        if dated.date is not None:
            encoded['date'] = dated.date

    def _encode_date(self, date: Date) -> Dict:
        encoded = {}
        if date.year:
            encoded['year'] = date.year
        if date.month:
            encoded['month'] = date.month
        if date.day:
            encoded['day'] = date.day
        return encoded

    def _encode_date_range(self, date: DateRange) -> Dict:
        encoded = {}
        if date.start:
            encoded['start'] = date.start
        if date.end:
            encoded['end'] = date.end
        return encoded

    def _encode_localized(self, encoded: Dict, localized: Localized) -> None:
        if localized.locale is not None:
            encoded['locale'] = localized.locale

    def _encode_has_media_type(self, encoded: Dict, media: HasMediaType) -> None:
        if media.media_type is not None:
            encoded['mediaType'] = media.media_type

    def _encode_has_links(self, encoded: Dict, has_links: HasLinks) -> None:
        if 'links' not in encoded:
            encoded['links'] = []
        for link in has_links.links:
            encoded['links'].append(link)

    def _encode_link(self, link: Link) -> Dict:
        encoded = {
            'url': link.url,
        }
        if link.label is not None:
            encoded['label'] = link.label
        if link.relationship is not None:
            encoded['relationship'] = link.relationship
        self._encode_localized(encoded, link)
        self._encode_has_media_type(encoded, link)
        return encoded

    def _encode_has_citations(self, encoded: Dict, has_citations: HasCitations) -> None:
        encoded['citations'] = [self._generate_url(
            citation) for citation in has_citations.citations if isinstance(citation, Identifiable)]

    def _encode_coordinates(self, coordinates: Point) -> Dict:
        return {
            '@context': {
                'latitude': 'https://schema.org/latitude',
                'longitude': 'https://schema.org/longitude',
            },
            '@type': 'https://schema.org/GeoCoordinates',
            'latitude': coordinates.latitude,
            'longitude': coordinates.longitude,
        }

    def _encode_localized_name(self, name: PlaceName) -> Dict:
        encoded = {
            'name': name.name,
        }
        self._encode_localized(encoded, name)
        return encoded

    def _encode_place(self, place: Place) -> Dict:
        encoded = {
            '@context': {
                'events': 'https://schema.org/event',
                'enclosedBy': 'https://schema.org/containedInPlace',
                'encloses': 'https://schema.org/containsPlace',
            },
            '@type': 'https://schema.org/Place',
            'id': place.id,
            'names': place.names,
            'events': [self._generate_url(event) for event in place.events],
            'enclosedBy': [self._generate_url(enclosure.enclosed_by) for enclosure in place.enclosed_by],
            'encloses': [self._generate_url(enclosure.encloses) for enclosure in place.encloses],
        }
        self._encode_schema(encoded, 'place')
        self._encode_identifiable_resource(encoded, place)
        self._encode_has_links(encoded, place)
        if place.coordinates is not None:
            encoded['coordinates'] = place.coordinates
            encoded['@context']['coordinates'] = 'https://schema.org/geo'
        return encoded

    def _encode_person(self, person: Person) -> Dict:
        encoded = {
            '@context': {
                'parents': 'https://schema.org/parent',
                'children': 'https://schema.org/child',
                'siblings': 'https://schema.org/sibling',
            },
            '@type': 'https://schema.org/Person',
            'id': person.id,
            'names': list(person.names),
            'parents': [self._generate_url(parent) for parent in person.parents],
            'children': [self._generate_url(child) for child in person.children],
            'siblings': [self._generate_url(sibling) for sibling in person.siblings],
            'private': person.private,
            'presences': [],
        }
        for presence in person.presences:
            if isinstance(presence.event, Identifiable):
                encoded['presences'].append({
                    '@context': {
                        'event': 'https://schema.org/performerIn',
                    },
                    'role': presence.role,
                    'event': self._generate_url(presence.event),
                })
        self._encode_schema(encoded, 'person')
        self._encode_identifiable_resource(encoded, person)
        self._encode_has_citations(encoded, person)
        self._encode_has_links(encoded, person)
        return encoded

    def _encode_person_name(self, name: PersonName) -> Dict:
        encoded = {}
        if name.individual is not None or name.affiliation is not None:
            encoded.update({
                '@context': {},
            })
        if name.individual is not None:
            encoded['@context']['individual'] = 'https://schema.org/givenName'
            encoded['individual'] = name.individual
        if name.affiliation is not None:
            encoded['@context']['affiliation'] = 'https://schema.org/familyName'
            encoded['affiliation'] = name.affiliation
        return encoded

    def _encode_file(self, file: File) -> Dict:
        encoded = {
            'id': file.id,
            'resources': [self._generate_url(entity) for entity in file.resources],
            'notes': file.notes,
        }
        self._encode_schema(encoded, 'file')
        self._encode_identifiable_resource(encoded, file)
        self._encode_has_media_type(encoded, file)
        return encoded

    def _encode_event(self, event: Event) -> Dict:
        encoded = {
            '@type': 'https://schema.org/Event',
            'type': event.type,
            'presences': [{
                '@context': {
                    'person': 'https://schema.org/actor',
                },
                'role': presence.role,
                'person': self._generate_url(presence.person),
            } for presence in event.presences],
        }
        self._encode_schema(encoded, 'event')
        self._encode_dated(encoded, event)
        self._encode_has_citations(encoded, event)
        if event.place is not None:
            encoded['place'] = self._generate_url(event.place)
            encoded.update({
                '@context': {},
            })
            encoded['@context']['place'] = 'https://schema.org/location'
        return encoded

    def _encode_identifiable_event(self, event: Event) -> Dict:
        encoded = self._encode_event(event)
        encoded['id'] = event.id
        self._encode_identifiable_resource(encoded, event)
        return encoded

    def _encode_event_type(self, event_type: EventType) -> str:
        return event_type.name()

    def _encode_presence_role(self, role: PresenceRole) -> str:
        return role.name()

    def _encode_identifiable_citation(self, citation: IdentifiableCitation) -> Dict:
        encoded = {
            '@type': 'https://schema.org/Thing',
            'id': citation.id,
            'facts': []
        }
        if isinstance(citation.source, Identifiable):
            encoded['source'] = self._generate_url(citation.source)
        for fact in citation.facts:
            if isinstance(fact, Identifiable):
                encoded['facts'].append(self._generate_url(fact))
        self._encode_schema(encoded, 'citation')
        self._encode_identifiable_resource(encoded, citation)
        self._encode_dated(encoded, citation)
        return encoded

    def _encode_identifiable_source(self, source: IdentifiableSource) -> Dict:
        encoded = {
            '@context': {
                'name': 'https://schema.org/name',
            },
            '@type': 'https://schema.org/Thing',
            'id': source.id,
            'name': source.name,
            'contains': [self._generate_url(contained) for contained in source.contains if isinstance(contained, Identifiable)],
            'citations': [self._generate_url(citation) for citation in source.citations if isinstance(citation, Identifiable)],
        }
        if source.author is not None:
            encoded['author'] = source.author
        if source.publisher is not None:
            encoded['publisher'] = source.publisher
        self._encode_schema(encoded, 'source')
        self._encode_identifiable_resource(encoded, source)
        self._encode_dated(encoded, source)
        self._encode_has_links(encoded, source)
        if source.contained_by is not None:
            encoded['containedBy'] = self._generate_url(source.contained_by)
        return encoded

    def _encode_note(self, note: Note) -> Dict:
        return {
            'text': note.text,
        }

    def _encode_media_type(self, media_type: MediaType) -> str:
        return str(media_type)
